//---------------------------------------------------------------------------
// Burn - Arcade emulator library
// Copyright (c) 2001 Dave (www.finalburn.com), all rights reserved.

// This refers to all the code except where stated otherwise
// (e.g. ym2151 emulator)

// You can use, modify and redistribute this code freely as long as you
// don't do so commercially. This copyright notice must remain with the code.
// If your program uses this code, you must either distribute or link to the
// source code. If you modify or improve this code, you must distribute
// the source code improvements.

// Dave
// Homepage: www.finalburn.com
// E-mail:  dave@finalburn.com
//---------------------------------------------------------------------------

#ifdef __cplusplus
 extern "C" {
#endif

#include <time.h>
#include "cheat.h"

extern unsigned int nCurrentFrame;

inline static int GetCurrentFrame() {
	return nCurrentFrame;
}

inline static void SetCurrentFrame(const unsigned int n) {
	nCurrentFrame = n;
}

extern bool bBurnUseMMX;

extern unsigned int nFramesEmulated;
extern unsigned int nFramesRendered;
extern clock_t starttime;					// system time when emulation started and after roms loaded

extern bool bForce60Hz;

extern int nBurnFPS;
extern int nBurnCPUSpeedAdjust;

#define BIT_DIGITAL			1

#define BIT_GROUP_ANALOG	4
#define BIT_ANALOG_REL		4
#define BIT_ANALOG_ABS		5

#define BIT_GROUP_CONSTANT	8
#define BIT_CONSTANT		8
#define BIT_DIPSWITCH		9

struct BurnRomInfo {
	char szName[16];
	unsigned int nLen;
	unsigned int nCrc;
	unsigned int nType;
};
struct BurnInputInfo {
	char* szName;
	unsigned char nType;
	union {
		unsigned char* pVal;					// Most inputs use a char*
		unsigned short* pShortVal;				// All analog inputs use a short*
	};
	char* szInfo;
};
struct BurnDIPInfo {
	int nInput;
	unsigned char nFlags;
	unsigned char nMask;
	unsigned char nSetting;
	char* szText;
};

struct BurnArea {
	void *Data;
	unsigned int nLen;
	int nAddress;
	char *szName;
};

extern int nBurnVer;						// Version number of the library

// burn.cpp
extern int (__cdecl *dprintf) (char* szFormat,...);
extern unsigned int nBurnDrvCount;			// Count of game drivers
extern unsigned int nBurnDrvSelect;			// Which game driver is selected

extern int nMaxPlayers;

extern unsigned char *pBurnDraw;			// Pointer to correctly sized bitmap
extern int nBurnPitch;						// Pitch between each line
extern int nBurnBpp;						// Bytes per pixel (2, 3, or 4)

extern unsigned char nBurnLayer;			// Can be used externally to select which layers to show
extern unsigned char nSpriteEnable;			// Can be used externally to select which Sprites to show

int BurnLibInit();
int BurnLibExit();
extern int (*BurnExtInput) (unsigned int i);// App-defined function to get the state of each input
char* BurnDrvText(unsigned int i);
int BurnDrvGetZipName(char** pszName, unsigned int i);
int BurnDrvGetRomInfo(struct BurnRomInfo *pri, unsigned int i);
int BurnDrvGetRomName(char** pszName, unsigned int i, int nAka);
int BurnDrvGetInputInfo(struct BurnInputInfo* pii, unsigned int i);
int BurnDrvGetDIPInfo(struct BurnDIPInfo* pdi, unsigned int i);
int BurnDrvGetVisibleSize(int* pnWidth, int* pnHeight);
int BurnDrvGetVisibleOffs(int* pnLeft, int* pnTop);
int BurnDrvGetFullSize(int* pnWidth, int* pnHeight);
int BurnDrvGetAspect(int* pnXAspect, int* pnYAspect);
int BurnDrvGetHardwareCode();
int BurnDrvGetFlags();
bool BurnDrvIsWorking();
int BurnDrvGetMaxPlayers();
int BurnDrvInit();
int BurnDrvExit();
int BurnDrvFrame();
int BurnDrvRedraw();
int BurnRecalcPal();

int BurnAreaScan(int nAction, int* pnMin);		// &1=for reading &2=for writing &4=Volatile &8=Non-Volatile &16=memcard
#define ACB_READ		 1
#define ACB_WRITE		 2

#define ACB_VOLATILE	 4
#define ACB_NVRAM		 8
#define ACB_MEMCARD		16
#define ACB_FULLSCAN	(ACB_VOLATILE | ACB_NVRAM | ACB_MEMCARD)

#define ACB_ACCESSMASK	(ACB_READ | ACB_WRITE)
#define ACB_TYPEMASK	(ACB_VOLATILE | ACB_NVRAM | ACB_MEMCARD)

extern unsigned int (*BurnHighCol) (int r, int g, int b, int i);
extern int (*BurnAcb) (struct BurnArea* pba);	// Area callback

// Progress bar
extern int (__cdecl *BurnExtProgressRangeCallback)(double fProgressRange);
extern int (__cdecl *BurnExtProgressUpdateCallback)(double fProgress, const char* pszText, bool bAbs);

int BurnSetProgressRange(double fProgressRange);
int BurnUpdateProgress(double fProgressStep, const char* pszText, bool bAbs);

// load.cpp
// Application-defined rom loading function:
extern int (*BurnExtLoadRom)(unsigned char* Dest, int* pnWrote, int i);

// state.cpp
int BurnStateLoadEmbed(FILE* fp, int nOffset, int bAll, int (*pLoadGame)());
int BurnStateLoad(char* szName, int bAll, int (*pLoadGame)());
int BurnStateSaveEmbed(FILE* fp, int nOffset, int bAll);
int BurnStateSave(char* szName, int bAll);

int BurnStateMAMEScan(int nAction, int* pnMin);
void BurnStateExit();
int BurnStateInit();


// burn.cpp
extern int nBurnSoundRate;					// Samplerate of sound
extern int nBurnSoundLen;					// Length in samples per frame
extern short* pBurnSoundOut;				// Pointer to output buffer

extern int nInterpolation;					// Desired interpolation level for ADPCM/PCM sound
extern int nFMInterpolation;				// Desired interpolation level for FM sound

// Flags used with the Burndriver structure

// Flags for the flags member
#define BDF_GAME_WORKING			(1 << 0)
#define BDF_ORIENTATION_FLIPPED		(1 << 1)
#define BDF_ORIENTATION_VERTICAL	(1 << 2)
#define BDF_BOARDROM				(1 << 3)
#define BDF_CLONE					(1 << 4)
#define BDF_BOOTLEG					(1 << 5)
#define BDF_PROTOTYPE				(1 << 6)
#define BDF_16BIT_ONLY				(1 << 7)

// Flags for the hardware member
// Format: 0xDDEEFFFF, where EE: Manufacturer, DD: Hardware platform, FFFF: Flags (used by driver)

#define HARDWARE_PUBLIC_MASK		0xFFFF0000
#define HARDWARE_PREFIX_MISC		0x00000000
#define HARDWARE_PREFIX_CAPCOM		0x01000000
#define HARDWARE_PREFIX_SEGA		0x02000000
#define HARDWARE_PREFIX_KONAMI		0x03000000
#define HARDWARE_PREFIX_TOAPLAN		0x04000000
#define HARDWARE_PREFIX_SNK			0x05000000
#define HARDWARE_PREFIX_CAVE		0x06000000

#define HARDWARE_MISC_MISC			HARDWARE_PREFIX_MISC

#define HARDWARE_CAPCOM_CPS1		(HARDWARE_PREFIX_CAPCOM | 0x00010000)
#define HARDWARE_CAPCOM_CPS1_QSOUND (HARDWARE_PREFIX_CAPCOM | 0x00020000)
#define HARDWARE_CAPCOM_CPS1_GENERIC (HARDWARE_PREFIX_CAPCOM | 0x00030000)
#define HARDWARE_CAPCOM_CPSCHANGER	(HARDWARE_PREFIX_CAPCOM | 0x00040000)
#define HARDWARE_CAPCOM_CPS2		(HARDWARE_PREFIX_CAPCOM | 0x00050000)
#define HARDWARE_CAPCOM_CPS2_SIMM	0x0002

#define HARDWARE_SEGA_SYSTEMX		(HARDWARE_PREFIX_SEGA | 0x00010000)
#define HARDWARE_SEGA_SYSTEMY		(HARDWARE_PREFIX_SEGA | 0x00020000)
#define HARDWARE_SEGA_SYSTEM16A		(HARDWARE_PREFIX_SEGA | 0x00030000)
#define HARDWARE_SEGA_SYSTEM16B 	(HARDWARE_PREFIX_SEGA | 0x00040000)
#define HARDWARE_SEGA_SYSTEM16M		(HARDWARE_PREFIX_SEGA | 0x00050000)
#define HARDWARE_SEGA_SYSTEM18		(HARDWARE_PREFIX_SEGA | 0x00060000)
#define HARDWARE_SEGA_SHARRIER		(HARDWARE_PREFIX_SEGA | 0x00070000)
#define HARDWARE_SEGA_OUTRUN		(HARDWARE_PREFIX_SEGA | 0x00080000)

#define HARDWARE_TOAPLAN_RAIZING	(HARDWARE_PREFIX_TOAPLAN | 0x00010000)
#define HARDWARE_TOAPLAN_68K_Zx80	(HARDWARE_PREFIX_TOAPLAN | 0x00020000)
#define HARDWARE_TOAPLAN_68K_ONLY	(HARDWARE_PREFIX_TOAPLAN | 0x00030000)

#define HARDWARE_SNK_NEOGEO			(HARDWARE_PREFIX_SNK | 0x00010000)
#define HARDWARE_SNK_SRAM			0x0002	// SRAM protection
#define HARDWARE_SNK_SWAPP			0x0004	// Swap code roms
#define HARDWARE_SNK_SWAPV			0x0008	// Swap sound roms
#define HARDWARE_SNK_SWAPC			0x0010	// Swap sprite roms
#define HARDWARE_SNK_ENCRYPTED_A	0x0020	// KOF99 encryption scheme
#define HARDWARE_SNK_ENCRYPTED_B	0x0040	// KOF2000 encryption scheme
#define HARDWARE_SNK_ALTERNATE_TEXT	0x0080	// KOF2000 text layer banks
#define HARDWARE_SNK_SMA_PROTECTION	0x0100  // SMA protection

#define HARDWARE_SNK_CONTROLMASK	0xF000
#define HARDWARE_SNK_JOYSTICK		0x0000  // Uses joysticks
#define HARDWARE_SNK_PADDLE			0x1000  // Uses joysticks or paddles
#define HARDWARE_SNK_TRACKBALL		0x2000  // Uses a trackball
#define HARDWARE_SNK_4_JOYSTICKS	0x3000	// Uses 4 joysticks
#define HARDWARE_SNK_MAHJONG		0x4000	// Uses 4 joysticks
#define HARDWARE_SNK_GAMBLING		0x5000  // Uses gambling controls

#define HARDWARE_CAVE_68K_ONLY		HARDWARE_PREFIX_CAVE
#define HARDWARE_CAVE_68K_Z80		(HARDWARE_PREFIX_CAVE | 0x0001)
#define HARDWARE_CAVE_M6295			0x0002
#define HARDWARE_CAVE_YM2151		0x0004

#ifdef __cplusplus
 } // End of extern "C"
#endif

