#include "burnint.h"
#include "burn_sound.h"
#include "burn_ym2610.h"

void (*BurnYM2610Render)(short* pSoundBuf, int nSegmentLength);

static int nBurnYM2610SoundRate;

static short* pBuffer;
static short* pYM2610Buffer[5];

static int* pAYBuffer;

static int nBurnPosition;
static unsigned int nSampleSize;
static unsigned int nFractionalPosition;
static unsigned int nSamplesRendered;

static void YM2610RenderResample(short* pSoundBuf, int nSegmentLength)
{
	unsigned int nSamplesNeeded;

	nBurnPosition += nSegmentLength;

	if (nBurnPosition >= nBurnSoundRate) {
		nBurnPosition = nSegmentLength;

		pYM2610Buffer[0][1] = pYM2610Buffer[0][(nFractionalPosition >> 16) - 3];
		pYM2610Buffer[0][2] = pYM2610Buffer[0][(nFractionalPosition >> 16) - 2];
		pYM2610Buffer[0][3] = pYM2610Buffer[0][(nFractionalPosition >> 16) - 1];

		pYM2610Buffer[1][1] = pYM2610Buffer[1][(nFractionalPosition >> 16) - 3];
		pYM2610Buffer[1][2] = pYM2610Buffer[1][(nFractionalPosition >> 16) - 2];
		pYM2610Buffer[1][3] = pYM2610Buffer[1][(nFractionalPosition >> 16) - 1];

		pYM2610Buffer[2][1] = pYM2610Buffer[2][(nFractionalPosition >> 16) - 3];
		pYM2610Buffer[2][2] = pYM2610Buffer[2][(nFractionalPosition >> 16) - 2];
		pYM2610Buffer[2][3] = pYM2610Buffer[2][(nFractionalPosition >> 16) - 1];

		nSamplesRendered -= (nFractionalPosition >> 16) - 4;

		for (unsigned int i = 0; i <= nSamplesRendered; i++) {
			pYM2610Buffer[0][4 + i] = pYM2610Buffer[0][(nFractionalPosition >> 16) + i];
			pYM2610Buffer[1][4 + i] = pYM2610Buffer[1][(nFractionalPosition >> 16) + i];
			pYM2610Buffer[2][4 + i] = pYM2610Buffer[2][(nFractionalPosition >> 16) + i];
		}

		nFractionalPosition &= 0x0000FFFF;
		nFractionalPosition |= 4 << 16;
	}

	pYM2610Buffer[0] = pBuffer + 4 + nSamplesRendered + 0 * 65536;
	pYM2610Buffer[1] = pBuffer + 4 + nSamplesRendered + 1 * 65536;
	pYM2610Buffer[2] = pBuffer + 4 + nSamplesRendered + 2 * 65536;
	pYM2610Buffer[3] = pBuffer + 4 + nSamplesRendered + 3 * 65536;
	pYM2610Buffer[4] = pBuffer + 4 + nSamplesRendered + 4 * 65536;

	nSamplesNeeded = (unsigned int)(nBurnPosition + 1) * nBurnYM2610SoundRate / nBurnSoundRate - nSamplesRendered;

	YM2610UpdateOne(0, &pYM2610Buffer[0], nSamplesNeeded);
	AY8910Update(0, &pYM2610Buffer[2], nSamplesNeeded);

	for (unsigned int i = 0; i < nSamplesNeeded; i++) {
		pYM2610Buffer[2][i] = (((int)pYM2610Buffer[2][i] + pYM2610Buffer[3][i] + pYM2610Buffer[4][i]) * (4096 * 60 / 100)) >> 12;
	}

	nSamplesRendered += nSamplesNeeded;

	pYM2610Buffer[0] = pBuffer + 0 * 65536;
	pYM2610Buffer[1] = pBuffer + 1 * 65536;
	pYM2610Buffer[2] = pBuffer + 2 * 65536;

	nSegmentLength <<= 1;

	for (int i = 0; i < nSegmentLength; i += 2, nFractionalPosition += nSampleSize) {
		int nSample;

#define CLIP(A) ((A) < -0x8000 ? -0x8000 : (A) > 0x7fff ? 0x7fff : (A))

		// Left channel
		nSample = INTERPOLATE4PS_16BIT((nFractionalPosition >> 4) & 0x0FFF,
									   pYM2610Buffer[0][(nFractionalPosition >> 16) - 3] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 3],
									   pYM2610Buffer[0][(nFractionalPosition >> 16) - 2] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 2],
									   pYM2610Buffer[0][(nFractionalPosition >> 16) - 1] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 1],
									   pYM2610Buffer[0][(nFractionalPosition >> 16) - 0] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 0]);
		pSoundBuf[i + 0] = CLIP(nSample);

		// Right channel
		nSample = INTERPOLATE4PS_16BIT((nFractionalPosition >> 4) & 0x0FFF,
									   pYM2610Buffer[1][(nFractionalPosition >> 16) - 3] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 3],
									   pYM2610Buffer[1][(nFractionalPosition >> 16) - 2] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 2],
									   pYM2610Buffer[1][(nFractionalPosition >> 16) - 1] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 1],
									   pYM2610Buffer[1][(nFractionalPosition >> 16) - 0] + pYM2610Buffer[2][(nFractionalPosition >> 16) - 0]);
		pSoundBuf[i + 1] = CLIP(nSample);

#undef CLIP

	}
}

static void YM2610RenderNormal(short* pSoundBuf, int nSegmentLength)
{
	nBurnPosition += nSegmentLength;

	pYM2610Buffer[0] = pBuffer + 0 * 8192;
	pYM2610Buffer[1] = pBuffer + 1 * 8192;
	pYM2610Buffer[2] = pBuffer + 2 * 8192;
	pYM2610Buffer[3] = pBuffer + 3 * 8192;
	pYM2610Buffer[4] = pBuffer + 4 * 8192;

	YM2610UpdateOne(0, &pYM2610Buffer[0], nSegmentLength);
	AY8910Update(0, &pYM2610Buffer[2], nSegmentLength);

	if (bBurnUseMMX) {
		for (int n = 0; n < nSegmentLength; n++) {
			pAYBuffer[n] = pYM2610Buffer[2][n] + pYM2610Buffer[3][n] + pYM2610Buffer[4][n];
		}
		BurnSoundCopy_FM_OPN_A(pYM2610Buffer[0], pAYBuffer, pSoundBuf, nSegmentLength, 65536 * 60 / 100, 65536 * 60 / 100);
	} else {
		for (int n = 0; n < nSegmentLength; n++) {
			int nAYSample, nTotalSample;

			nAYSample  = pYM2610Buffer[2][n];
			nAYSample += pYM2610Buffer[3][n];
			nAYSample += pYM2610Buffer[4][n];

			nAYSample  *= 4096 * 60 / 100;
			nAYSample >>= 12;

			nTotalSample = nAYSample + pYM2610Buffer[0][n];
			if (nTotalSample < -32768) {
				nTotalSample = -32768;
			} else {
				if (nTotalSample > 32767) {
					nTotalSample = 32767;
				}
			}
			pSoundBuf[(n << 1) + 0] = nTotalSample;

			nTotalSample = nAYSample + pYM2610Buffer[1][n];
			if (nTotalSample < -32768) {
				nTotalSample = -32768;
			} else {
				if (nTotalSample > 32767) {
					nTotalSample = 32767;
				}
			}
			pSoundBuf[(n << 1) + 1] = nTotalSample;
		}
	}
}

void BurnYM2610Reset()
{
	YM2610ResetChip(0);
}

void BurnYM2610Exit()
{
	YM2610Shutdown();
	AY8910Exit(0);

	BurnFMTimerExit();

	free(pBuffer);
	free(pAYBuffer);
}

int BurnYM2610Init(int nClockFrequency, unsigned char* YM2610ADPCMAROM, int* nYM2610ADPCMASize, unsigned char* YM2610ADPCMBROM, int* nYM2610ADPCMBSize, FM_IRQHANDLER IRQCallback)
{
	if (nBurnSoundRate <= 0) {
		AY8910InitYM(0, nClockFrequency, 11025, NULL, NULL, NULL, NULL);
		YM2610Init(1, nClockFrequency, 11025, (void**)(&YM2610ADPCMAROM), nYM2610ADPCMASize, (void**)(&YM2610ADPCMBROM), nYM2610ADPCMBSize, &BurnOPNTimerCallback, IRQCallback);
		return 0;
	}

	if (nFMInterpolation == 3) {
		// Set YM2610 core samplerate to match the hardware
		nBurnYM2610SoundRate = nClockFrequency / 144;
		// Bring YM2610 core samplerate within usable range
		while (nBurnYM2610SoundRate > nBurnSoundRate * 3) {
			nBurnYM2610SoundRate >>= 1;
		}

		BurnYM2610Render = YM2610RenderResample;
	} else {
		nBurnYM2610SoundRate = nBurnSoundRate;

		BurnYM2610Render = YM2610RenderNormal;
	}

	BurnFMTimerInit(&YM2610TimerOver);

	AY8910InitYM(0, nClockFrequency, nBurnYM2610SoundRate, NULL, NULL, NULL, NULL);
	YM2610Init(1, nClockFrequency, nBurnYM2610SoundRate, (void**)(&YM2610ADPCMAROM), nYM2610ADPCMASize, (void**)(&YM2610ADPCMBROM), nYM2610ADPCMBSize, &BurnOPNTimerCallback, IRQCallback);

	pBuffer = (short*)malloc(65536 * 5 * sizeof(short));
	memset(pBuffer, 0, 65536 * 5 * sizeof(short));

	pAYBuffer = (int*)malloc(65536 * sizeof(int));
	memset(pAYBuffer, 0, 65536 * sizeof(int));

	nSampleSize = (unsigned int)nBurnYM2610SoundRate * (1 << 16) / nBurnSoundRate;
	nFractionalPosition = 4 << 16;
	nSamplesRendered = 0;
	nBurnPosition = 0;

	return 0;
}

void BurnYM2610Scan(int /*nAction*/)
{
}
