#include "cps.h"

/*==============================================
 Mars Matrix
 ==============================================*/

static struct BurnInputInfo DrvInputList[] =
{
  {"P1 Coin"       , BIT_DIGITAL, CpsInp020+4, "p1 coin"  },
  {"P1 Start"      , BIT_DIGITAL, CpsInp020+0, "p1 start" },
  {"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"    },
  {"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"  },
  {"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"  },
  {"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right" },
  {"P1 Shot"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},

  {"P2 Coin"       , BIT_DIGITAL, CpsInp020+5, "p2 coin"  },
  {"P2 Start"      , BIT_DIGITAL, CpsInp020+1, "p2 start" },
  {"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"    },
  {"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"  },
  {"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"  },
  {"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right" },
  {"P2 Shot"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},

  {"Reset"         , BIT_DIGITAL, &CpsReset  , "reset"    },
  {"Diagnostic"    , BIT_DIGITAL, CpsInp021+1, "diag"     },
  {"Service"       , BIT_DIGITAL, CpsInp021+2, "service"  },
};

STDINPUTINFO(Drv);


/*=================
 Mars Matrix (USA)
=================*/


static struct BurnRomInfo mmatrixRomDesc[]=
{
  {"mmxux.03" ,0x80000,0x7868AE77, 0x10},
  {"mmxux.04" ,0x80000,0xA5EE6D07, 0x10},
  {"mmxux.05" ,0x80000,0xB07745FF, 0x10},

  {"mmxu.03"  ,0x80000,0xAB65b599, 0x10},
  {"mmxu.04"  ,0x80000,0x0135FC6C, 0x10},
  {"mmxu.05"  ,0x80000,0xF1FD2B84, 0x10},

  {"mmx.13"  ,0x400000,0x04748718,    1},
  {"mmx.15"  ,0x400000,0x38074F44,    1},
  {"mmx.17"  ,0x400000,0xE4635E35,    1},
  {"mmx.19"  ,0x400000,0x4400A3F2,    1},
  {"mmx.14"  ,0x400000,0xD52BF491,    1},
  {"mmx.16"  ,0x400000,0x23F70780,    1},
  {"mmx.18"  ,0x400000,0x2562C9D5,    1},
  {"mmx.20"  ,0x400000,0x583A9687,    1},

  {"mmx.01"  ,0x020000,0xC57E8171,    2},

  {"mmx.11"  ,0x400000,0x4180B39F,    2},
  {"mmx.12"  ,0x400000,0x95E22A59,    2},
};

STD_ROM_PICK(mmatrix) STD_ROM_FN(mmatrix)

struct BurnDriver BurnDrvCpsMarsMatrix=
{
  {"mmatrix", "Mars Matrix: Hyper Solid Shooting (US 000412)", NULL, "Capcom / Takumi", "CPS2", "2000", NULL, NULL},
  BDF_GAME_WORKING, 2, HARDWARE_CAPCOM_CPS2,
  NULL, mmatrixRomInfo, mmatrixRomName, DrvInputInfo, NULL,
  Cps2Init, CpsExit,Cps2Frame,CpsRedraw,CpsAreaScan,
  &CpsRecalcPal,384,224,4,3
};

/*===================
 Mars Matrix (Japan)
===================*/


static struct BurnRomInfo mmatrixjRomDesc[]=
{
	{"mmxjx.03" ,0x80000,0x4CA1424F, 0x10},
	{"mmxjx.04" ,0x80000,0x61B9B2A1, 0x10},
	{"mmxjx.05" ,0x80000,0xBDD304CF, 0x10},

	{"mmxj.03"  ,0x80000,0x1D5DE213, 0x10},
	{"mmxj.04"  ,0x80000,0xD943A339, 0x10},
	{"mmxj.05"  ,0x80000,0x0C8B4ABB, 0x10},

	{"mmx.13"  ,0x400000,0x04748718,    1},
	{"mmx.15"  ,0x400000,0x38074F44,    1},
	{"mmx.17"  ,0x400000,0xE4635E35,    1},
	{"mmx.19"  ,0x400000,0x4400A3F2,    1},
	{"mmx.14"  ,0x400000,0xD52BF491,    1},
	{"mmx.16"  ,0x400000,0x23F70780,    1},
	{"mmx.18"  ,0x400000,0x2562C9D5,    1},
	{"mmx.20"  ,0x400000,0x583A9687,    1},

	{"mmx.01"  ,0x020000,0xC57E8171,    2},

	{"mmx.11"  ,0x400000,0x4180B39F,    2},
	{"mmx.12"  ,0x400000,0x95E22A59,    2},
};

STD_ROM_PICK(mmatrixj) STD_ROM_FN(mmatrixj)

struct BurnDriver BurnDrvCpsMarsMatrixJ=
{
  {"mmatrixj", "Mars Matrix: Hyper Solid Shooting (Japan 000412)", NULL, "Capcom / Takumi", "CPS2", "2000", "mmatrix", NULL},
  BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS2,
  NULL, mmatrixjRomInfo, mmatrixjRomName, DrvInputInfo, NULL,
  Cps2Init, CpsExit,Cps2Frame,CpsRedraw,CpsAreaScan,
  &CpsRecalcPal,384,224,4,3
};

