// FM timers
#include "burnint.h"
#include "timer.h"

double dTime;									// Time elapsed since the emulated machine was started

static bool bTimerARunning, bTimerBRunning;
static unsigned int nTimerACount, nTimerBCount;
static unsigned int nTimerA, nTimerB;

static int (*pFMTimerOverCallback)(int, int);

extern "C" double BurnFMTimerGetTime()
{
	return dTime;
}

int BurnFMTimerUpdate(unsigned int nTicks)
{
	int nIRQStatus = 0;

	if (bTimerARunning) {

		nTimerA	+= nTicks;

		while (nTimerA > nTimerACount) {
			nTimerA -= nTimerACount;
			nIRQStatus |= pFMTimerOverCallback(0, 0);
		}
	}

	if (bTimerBRunning) {

		nTimerB	+= nTicks;

		while (nTimerB > nTimerBCount) {
			nTimerB -= nTimerBCount;
			nIRQStatus |= pFMTimerOverCallback(0, 1);
		}
	}

	return nIRQStatus;
}

void BurnOPLTimerCallback(int c, double period)
{
	if (c == 0) {							// Timer A
		if (period == 0.0) {
			nTimerA = 0;
			bTimerARunning = false;
			return;
		}

		nTimerACount = (unsigned int)(period * (double)TIMER_TICKS_PER_SECOND);
		bTimerARunning = true;
	} else {								// Timer B
		if (period == 0.0) {
			nTimerB = 0;
			bTimerBRunning = false;
			return;
		}

		nTimerBCount = (unsigned int)(period * (double)TIMER_TICKS_PER_SECOND);
		bTimerBRunning = true;
	}
}

void BurnOPNTimerCallback(int /* n */, int c, int cnt, double stepTime)
{
	if (c == 0) {							// Timer A
		if (cnt == 0) {
			nTimerA = 0;
			bTimerARunning = false;
			return;
		}

		nTimerACount = (unsigned int)(stepTime * (double)TIMER_TICKS_PER_SECOND);
		nTimerACount *= cnt;

		bTimerARunning = true;
	} else {								// Timer B
		if (cnt == 0) {
			nTimerB = 0;
			bTimerBRunning = false;
			return;
		}

		nTimerBCount = (unsigned int)(stepTime * (double)TIMER_TICKS_PER_SECOND);
		nTimerBCount *= cnt;

		bTimerBRunning = true;
	}
}

void BurnFMTimerScan(int nAction)
{
	if (nAction & ACB_VOLATILE) {
		SCAN_VAR(nTimerA); SCAN_VAR(nTimerACount); SCAN_VAR(bTimerARunning);
		SCAN_VAR(nTimerB); SCAN_VAR(nTimerBCount); SCAN_VAR(bTimerBRunning);
		SCAN_VAR(dTime);
	}
}

void BurnFMTimerExit()
{
	return;
}

int BurnFMTimerInit(int (*pCallback)(int, int)) {

	pFMTimerOverCallback = pCallback;

	nTimerA = nTimerB = 0;

	bTimerARunning = bTimerBRunning = false;

	dTime = 0.0;

	return 0;
}
