#include "toaplan.h"
#include "ymz280b.h"
#include "eeprom_93cxx.h"

// Battle Bakraid.

static unsigned char nEEPROMData[0x42] = {
																						0xC2, 0x49,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x33, 0xCF, 0x00, 0x04, 0x00, 0x00, 0x27, 0x14,
	0x00, 0x05, 0x06, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC2, 0x49,
};

static unsigned char DrvButton[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static unsigned char DrvJoy1[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static unsigned char DrvJoy2[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static unsigned char DrvInput[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

static unsigned char DrvReset = 0;
static unsigned char bDrawScreen;
static bool bVBlank;

static unsigned char nIRQPending;

static int nSoundData[4];

static int Z80BusRQ = 0;
static int nZ80IRQCycles = 0;
static const int nZ80IRQPeriod = TOA_Z80_SPEED / 445;

static int nTextROMStatus;

static struct BurnRomInfo bbakraidRomDesc[] = {
	{"prg0u022.new", 0x080000, 0xFA8D38D3, 0x10}, //  0 CPU #0 code (even)
	{"prg2u021.bin", 0x080000, 0xFFBA8656, 0x10}, //  1
	{"prg1u023.new", 0x080000, 0x4AE9AA64, 0x10}, //  2				(odd)
	{"prg3u024.bin", 0x080000, 0x834B8AD6, 0x10}, //  3

	{"gfxu0510.bin", 0x400000, 0x9CCA3446,    1}, //  4 GP9001 Tile data
	{"gfxu0512.bin", 0x400000, 0xA2A281D5,    1}, //  5
	{"gfxu0511.bin", 0x400000, 0xE16472C0,    1}, //  6
	{"gfxu0513.bin", 0x400000, 0x8BB635A0,    1}, //  7

	{"sndu0720.bin", 0x020000, 0xE62AB246, 0x10}, //  8 Z80 program

	{"rom6.829",	0x400000, 0x8848B4A0,    2}, //  9 YMZ280B (AD)PCM data
	{"rom7.830",	0x400000, 0xD6224267,    2}, // 10
	{"rom8.831",	0x400000, 0xA101DFB0,    2}, // 11
};

STD_ROM_PICK(bbakraid);
STD_ROM_FN(bbakraid);

// Rom information
static struct BurnRomInfo bbakradaRomDesc[] = {
	{"prg0u022.bin", 0x080000, 0x0DD59512, 0x10}, //  0 CPU #0 code (even)
	{"prg2u021.bin", 0x080000, 0xFFBA8656, 0x10}, //  1
	{"prg1u023.bin", 0x080000, 0xFECDE223, 0x10}, //  2				(odd)
	{"prg3u024.bin", 0x080000, 0x834B8AD6, 0x10}, //  3

	{"gfxu0510.bin", 0x400000, 0x9CCA3446,    1}, //  4 GP9001 Tile data
	{"gfxu0512.bin", 0x400000, 0xA2A281D5,    1}, //  5
	{"gfxu0511.bin", 0x400000, 0xE16472C0,    1}, //  6
	{"gfxu0513.bin", 0x400000, 0x8BB635A0,    1}, //  7

	{"sndu0720.bin", 0x020000, 0xE62AB246, 0x10}, //  8 Z80 program

	{"rom6.829",	0x400000, 0x8848B4A0,    2}, //  9 YMZ280B (AD)PCM data
	{"rom7.830",	0x400000, 0xD6224267,    2}, // 10
	{"rom8.831",	0x400000, 0xA101DFB0,    2}, // 11
};

STD_ROM_PICK(bbakrada);
STD_ROM_FN(bbakrada);

static struct BurnInputInfo bbakraidInputList[] = {
	{"P1 Coin",		BIT_DIGITAL,	DrvButton + 3,	"p1 coin"},
	{"P1 Start",	BIT_DIGITAL,	DrvButton + 5,	"p1 start"},

	{"P1 Up",		BIT_DIGITAL,	DrvJoy1 + 0,	"p1 up"},
	{"P1 Down",		BIT_DIGITAL,	DrvJoy1 + 1,	"p1 down"},
	{"P1 Left",		BIT_DIGITAL,	DrvJoy1 + 2,	"p1 left"},
	{"P1 Right",	BIT_DIGITAL,	DrvJoy1 + 3,	"p1 right"},
	{"P1 Button 1",	BIT_DIGITAL,	DrvJoy1 + 4,	"p1 fire 1"},
	{"P1 Button 2",	BIT_DIGITAL,	DrvJoy1 + 5,	"p1 fire 2"},
	{"P1 Button 3",	BIT_DIGITAL,	DrvJoy1 + 6,	"p1 fire 3"},

	{"P2 Coin",		BIT_DIGITAL,	DrvButton + 4,	"p2 coin"},
	{"P2 Start",	BIT_DIGITAL,	DrvButton + 6,	"p2 start"},

	{"P2 Up",		BIT_DIGITAL,	DrvJoy2 + 0,	"p2 up"},
	{"P2 Down",		BIT_DIGITAL,	DrvJoy2 + 1,	"p2 down"},
	{"P2 Left",		BIT_DIGITAL,	DrvJoy2 + 2,	"p2 left"},
	{"P2 Right",	BIT_DIGITAL,	DrvJoy2 + 3,	"p2 right"},
	{"P2 Button 1",	BIT_DIGITAL,	DrvJoy2 + 4,	"p2 fire 1"},
	{"P2 Button 2",	BIT_DIGITAL,	DrvJoy2 + 5,	"p2 fire 2"},
	{"P2 Button 3",	BIT_DIGITAL,	DrvJoy2 + 6,	"p2 fire 3"},

	{"Reset",		BIT_DIGITAL,	&DrvReset,		"reset"},
	{"Diagnostics",	BIT_DIGITAL,	DrvButton + 0,	"diag"},
	{"Dip A",		BIT_DIPSWITCH,	DrvInput + 3,	"dip"},
	{"Dip B",		BIT_DIPSWITCH,	DrvInput + 4,	"dip"},
	{"Dip C",		BIT_DIPSWITCH,	DrvInput + 5,	"dip"},
};

STDINPUTINFO(bbakraid);

static struct BurnDIPInfo bbakraidDIPList[] = {

	// Defaults
	{0x14,	0xFF, 0xFF,	0x00, NULL},
	{0x15,	0xFF, 0xFF,	0x00, NULL},
	{0x16,	0xFF, 0x3F,	0x00, NULL},

	// DIP 1
	{0,		0xFE, 0,	2,	  "Test mode"},
	{0x14,	0x01, 0x01,	0x00, "Normal"},
	{0x14,	0x01, 0x01,	0x01, "Test"},
	// Normal
	{0,		0xFE, 0,	2,	  "Starting coin"},
	{0x14,	0x82, 0x02,	0x00, "1 credit"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0x02,	0x02, "2 credits"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	// Free play
	{0,		0xFE, 0,	2,	  "Stick mode"},
	{0x14,	0x02, 0x02,	0x00, "Normal"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x02, 0x02,	0x02, "Special"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	// Normal
	{0,		0xFE, 0,	8,	  "Coin 1"},
	{0x14,	0x01, 0x1C, 0x00, "1 coin 1 credit"},
	{0x14,	0x01, 0x1C, 0x04, "1 coin 2 credits"},
	{0x14,	0x01, 0x1C, 0x08, "1 coin 3 credits"},
	{0x14,	0x01, 0x1C, 0x0C, "1 coin 3 credits"},
	{0x14,	0x01, 0x1C, 0x10, "2 coins 1 credit"},
	{0x14,	0x01, 0x1C, 0x14, "3 coins 1 credit"},
	{0x14,	0x01, 0x1C, 0x18, "4 coins 1 credit"},
	{0x14,	0x01, 0x1C, 0x1C, "Free Play"},
	// 0x1C: Free play settings active
	// Normal
	{0,		0xFE, 0,	7,	  "Coin 2"},
	{0x14,	0x82, 0xE0,	0x00, "1 coin 1 credit"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0x20, "1 coin 2 credits"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0x40, "1 coin 3 credits"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0x60, "1 coin 3 credits"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0x80, "2 coins 1 credit"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0xA0, "3 coins 1 credit"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0xC0, "4 coins 1 credit"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x82, 0xE0,	0x04, "1 coin 1 credit"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	// Free play
	{0,		0xFE, 0,	2,	  "Hit score"},
	{0x14,	0x02, 0x20, 0x00, "Off"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x02, 0x20, 0x20, "On"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0,		0xFE, 0,	2,	  "Sound effect"},
	{0x14,	0x02, 0x40,	0x00, "Off"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x02, 0x40,	0x40, "On"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0,		0xFE, 0,	2,	  "Music"},
	{0x14,	0x02, 0x80,	0x00, "Off"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},
	{0x14,	0x02, 0x80,	0x80, "On"},
	{0x14,	0x00, 0x1C, 0x1C, NULL},

	// DIP 2
	{0,		0xFE, 0,	4,	  "Start rank"},
	{0x15,	0x01, 0x03, 0x00, "Normal"},
	{0x15,	0x01, 0x03, 0x01, "Easy"},
	{0x15,	0x01, 0x03, 0x02, "Hard"},
	{0x15,	0x01, 0x03, 0x03, "Very hard"},
	{0,		0xFE, 0,	4,	  "Timer rank"},
	{0x15,	0x01, 0x0C, 0x00, "Normal"},
	{0x15,	0x01, 0x0C, 0x04, "Low"},
	{0x15,	0x01, 0x0C, 0x08, "High"},
	{0x15,	0x01, 0x0C, 0x0C, "Highest"},
	{0,		0xFE, 0,	4,	  "Player counts"},
	{0x15,	0x01, 0x30, 0x00, "3"},
	{0x15,	0x01, 0x30, 0x01, "4"},
	{0x15,	0x01, 0x30, 0x02, "2"},
	{0x15,	0x01, 0x30, 0x03, "1"},
	{0,		0xFE, 0,	4,	  "Extra player"},
    {0x15,	0x01, 0xC0, 0x00, "2000000 each"},
    {0x15,	0x01, 0xC0, 0x40, "3000000 each"},
    {0x15,	0x01, 0xC0, 0x80, "4000000 each"},
    {0x15,	0x01, 0xC0, 0xC0, "No extra player"},

	// DIP 3
	{0,		0xFE, 0,	2,	  "Screen flip"},
	{0x16,	0x01, 0x01, 0x00, "Off"},
	{0x16,	0x01, 0x01, 0x01, "On"},
	{0,		0xFE, 0,	2,	  "Demo sound"},
	{0x16,	0x01, 0x02, 0x00, "On"},
	{0x16,	0x01, 0x02, 0x02, "Off"},
	{0,		0xFE, 0,	2,	  "Stage edit"},
	{0x16,	0x01, 0x04, 0x00, "Disable"},
	{0x16,	0x01, 0x04, 0x04, "Enable"},
	{0,		0xFE, 0,	2,	  "Continue play"},
	{0x16,	0x01, 0x08, 0x00, "Enable"},
	{0x16,	0x01, 0x08, 0x08, "Disable"},
	{0,		0xFE, 0,	2,	  "Invincible"},
	{0x16,	0x01, 0x10, 0x00, "Off"},
	{0x16,	0x01, 0x10, 0x10, "On"},
	{0,		0xFE, 0,	2,	  "Score ranking"},
	{0x16,	0x01, 0x20, 0x00, "Save"},
	{0x16,	0x01, 0x20, 0x20, "No save"},
};

STDDIPINFO(bbakraid);

static unsigned char *Mem = NULL, *MemEnd = NULL;
static unsigned char *RamStart, *RamEnd;
static unsigned char *Rom01;
static unsigned char *Ram01, *Ram02, *RamPal;

static int nColCount = 0x0800;

static int MemIndex()
{
	unsigned char *Next; Next = Mem;
	Rom01		= Next; Next += 0x200000;			//
	RomZ80		= Next; Next += 0x020000;			// Z80 ROM
	GP9001ROM[0]= Next; Next += nGP9001ROMSize[0];	// GP9001 tile data
	YMZ280BROM	= Next;	Next += 0xC00000;
	RamStart	= Next;
	ExtraTROM	= Next; Next += 0x008000;			// Extra Text layer tile data
	ExtraTRAM	= Next; Next += 0x002000;			// Extra tile layer
	Ram01		= Next; Next += 0x005000;			// RAM + Extra text layer scroll/offset
	Ram02		= Next; Next += 0x008000;			//
	RamPal		= Next; Next += 0x001000;			// palette
	RamZ80		= Next; Next += 0x004000;			// Z80 RAM
	GP9001RAM[0]= Next; Next += 0x004000;
	GP9001Reg[0]= (unsigned short*)Next; Next += 0x0100 * sizeof(short);
	RamEnd		= Next;
	ToaPalette	= (unsigned int *)Next; Next += nColCount * sizeof(unsigned int);
	MemEnd		= Next;

 	ExtraTSelect= Ram01;							// Extra text layer scroll
	ExtraTScroll= Ram01 + 0x000200;					// Extra text layer offset
	return 0;
}

// Scan ram
static int DrvScan(int nAction, int *pnMin)
{
	struct BurnArea ba;

	if (pnMin) {						// Return minimum compatible version
		*pnMin = 0x029402;
	}

	EEPROMScan(nAction, pnMin);			// Scan EEPROM

	if (nAction & ACB_VOLATILE) {		// Scan volatile ram
		memset(&ba, 0, sizeof(ba));
    	ba.Data		= RamStart;
		ba.nLen		= RamEnd - RamStart;
		ba.szName	= "RAM";
		BurnAcb(&ba);

		SekScan(nAction);				// scan 68000 states
		ZetScan(nAction);				// Scan Z80

		YMZ280BScan();

		SCAN_VAR(GP9001TileBank);
		SCAN_VAR(DrvInput);
		SCAN_VAR(nSoundData);
		SCAN_VAR(Z80BusRQ);
		SCAN_VAR(nIRQPending);
	}

	return 0;
}

static int LoadRoms()
{
	// Load 68000 ROM
	if (ToaLoadCode(Rom01, 0, 4)) {
		return 1;
	}

	// Load GP9001 tile data
	ToaLoadGP9001Tiles(GP9001ROM[0], 4, 4, nGP9001ROMSize[0]);

	// Load the z80 Rom
	BurnLoadRom(RomZ80, 8, 1);

	// Load YMZ280B (AD)PCM data
	BurnLoadRom(YMZ280BROM + 0x000000, 9, 1);
	BurnLoadRom(YMZ280BROM + 0x400000, 10, 1);
	BurnLoadRom(YMZ280BROM + 0x800000, 11, 1);

	return 0;
}

unsigned char __fastcall bbakraidZIn(unsigned short nAddress)
{
	nAddress &= 0xFF;
	switch (nAddress) {
		case 0x48:
			return nSoundData[0];
		case 0x4A:
			return nSoundData[1];

		case 0x81:
			// YMZ280B status port.
			return YMZ280BReadStatus();

//		default:
//			printf("Z80 read port #%02X\n", nAddress & 0xFF);
	}
	return 0;
}

void __fastcall bbakraidZOut(unsigned short nAddress, unsigned char nValue)
{
	nAddress &= 0xFF;
	switch (nAddress) {
		case 0x40:
			nSoundData[2] = nValue;
			break;
		case 0x42:
			nSoundData[3] = nValue;
			break;

//		case 0x44:				// ???
//		case 0x46:				// Acknowledge interrupt
//			break;

		case 0x80:
			// YMZ280B register select
			YMZ280BSelectRegister(nValue);
			break;
		case 0x81:
			// YMZ280B register write
			YMZ280BWriteRegister(nValue);
			break;

//		default:
//			printf("Z80 wrote value %02X to port #%02x.\n", nValue, nAddress);
	}
}

static int DrvZ80Init()
{
	// Init the Z80
	ZetInit(1);

	ZetSetInHandler(bbakraidZIn);
	ZetSetOutHandler(bbakraidZOut);

	// ROM
	ZetMapArea(0x0000, 0xBFFF, 0, RomZ80);	// Direct Read from ROM
	ZetMapArea(0x0000, 0xBFFF, 2, RomZ80);	// Direct Fetch from ROM
	// RAM
	ZetMapArea(0xC000, 0xFFFF, 0, RamZ80);	// Direct Read from RAM
	ZetMapArea(0xC000, 0xFFFF, 1, RamZ80);	// Direct Write to RAM
	ZetMapArea(0xC000, 0xFFFF, 2, RamZ80);	//

	ZetMemEnd();

	return 0;
}

static void Map68KTextROM(bool bMapTextROM)
{
	if (bMapTextROM) {
		if (nTextROMStatus != 1) {
			SekMapMemory(ExtraTROM,  0x200000, 0x207FFF, SM_RAM);	// Extra text tile memory

			nTextROMStatus = 1;
		}
	} else {
		if (nTextROMStatus != 0) {
			SekMapMemory(ExtraTRAM,	0x200000, 0x201FFF, SM_RAM);	// Extra text tilemap RAM
			SekMapMemory(RamPal,	0x202000, 0x202FFF, SM_RAM);	// Palette RAM
			SekMapMemory(Ram01,		0x203000, 0x207FFF, SM_RAM);	// Extra text Scroll & offset; RAM

			nTextROMStatus = 0;
		}
	}
}

unsigned char __fastcall bbakraidReadByte(unsigned int sekAddress)
{
	switch (sekAddress) {

		case 0x500000:								// Player 2 Inputs
			return DrvInput[1];
		case 0x500001:								// Player 1 Inputs
			return DrvInput[0];
		case 0x500002:								// Other inputs
			return DrvInput[5];
		case 0x500003:								// Dipswitch 3
			return DrvInput[2];
		case 0x500004:								// Dipswitch 2
			return DrvInput[4];
		case 0x500005:								// Dipswitch 1
			return DrvInput[3];

		// These addresses contain the response of the Z80 to the (sound) commands
		case 0x500011:
			return nSoundData[2];
		case 0x500013:
			return nSoundData[3];

		case 0x500019:
			return (EEPROMRead() << 4) | (Z80BusRQ >> 4);

		default: {
//			printf("Attempt to read byte value of location %x\n", sekAddress);
		}
	}

	return 0;
}

unsigned short __fastcall bbakraidReadWord(unsigned int sekAddress)
{
	switch (sekAddress) {

		// Video status register
		case 0x500006:
			return ToaScanlineRegister();

		// These addresses contain the response of the Z80 to the (sound) commands
		case 0x500010:
			return nSoundData[2];
		case 0x500012:
			return nSoundData[3];

		case 0x500018:
			return (EEPROMRead() << 4) | (Z80BusRQ >> 4);

		default: {
//			printf("Attempt to read word value of location %x\n", sekAddress);
		}
	}

	return 0;
}

void __fastcall bbakraidWriteByte(unsigned int sekAddress, unsigned char byteValue)
{
	switch (sekAddress) {

		case 0x500009:								// Coin control
			return;

		case 0x50001F:
			Z80BusRQ = byteValue & 0x10;
			EEPROMWrite(byteValue & 8, byteValue & 1, byteValue & 4);
			return;

//		default:
//			printf("Attempt to write byte value %x to location %x\n", byteValue, sekAddress);
	}
}

void __fastcall bbakraidWriteWord(unsigned int sekAddress, unsigned short wordValue)
{
	switch (sekAddress) {
		case 0x500014:
			nSoundData[0] = wordValue;
			return;
		case 0x500016:
			nSoundData[1] = wordValue;
			return;

		// This register is always written to after writing (sound) commands for the Z80
		case 0x50001A:
			ZetNmi();
			nCyclesDone[1] += ZetRun(0x0400);
			return;

		// Serial EEPROM command
		case 0x50001E:
			Z80BusRQ = wordValue & 0x10;
			EEPROMWrite(wordValue & 8, wordValue & 1, wordValue & 4);
			return;

		case 0x500080:
			Map68KTextROM(false);
			return;

		case 0x50001C:		// ???
			return;

		case 0x500082:		// Acknowledge interrupt
			SekSetIRQLine(0, 0);
			nIRQPending = 0;
			return;

		case 0x5000C0:
		case 0x5000C1:
		case 0x5000C2:
		case 0x5000C3:
		case 0x5000C4:
		case 0x5000C5:
		case 0x5000C6:
		case 0x5000C7:
		case 0x5000C8:
		case 0x5000C9:
		case 0x5000CA:
		case 0x5000CB:
		case 0x5000CC:
		case 0x5000CD:
		case 0x5000CE:
			GP9001TileBank[(sekAddress & 0x0F) >> 1] = ((wordValue & 0x0F) << 15);
			return;

//		default:
//			printf("Attempt to write word value %x to location %x\n", wordValue, sekAddress);
	}
}

unsigned short __fastcall bbakraidReadWordGP9001(unsigned int sekAddress)
{
	switch (sekAddress) {
		case 0x400008:
			return ToaGP9001ReadRAM_Hi(0);
		case 0x40000A:
			return ToaGP9001ReadRAM_Lo(0);

	}

	return 0;
}

void __fastcall bbakraidWriteWordGP9001(unsigned int sekAddress, unsigned short wordValue)
{
	switch (sekAddress) {

		case 0x400000:
			ToaGP9001WriteRegister(wordValue);
			break;

		case 0x400004:
			ToaGP9001SelectRegister(wordValue);
			break;

		case 0x400008:
		case 0x40000A:
			ToaGP9001WriteRAM(wordValue, 0);
			break;

		case 0x40000C:								// Set GP9001 VRAM address-pointer
			ToaGP9001SetRAMPointer(wordValue);
			break;
	}
}

unsigned char __fastcall bbakraidReadByteZ80ROM(unsigned int sekAddress)
{
	return RomZ80[(sekAddress & 0x7FFFF) >> 1];
}

unsigned short __fastcall bbakraidReadWordZ80ROM(unsigned int sekAddress)
{
	return RomZ80[(sekAddress & 0x7FFFF) >> 1];
}

static int DrvDoReset()
{

	SekOpen(0);
	nIRQPending = 0;
    SekSetIRQLine(0, 0);
	SekReset();
	SekClose();

	ZetReset();

	EEPROMReset();

	Z80BusRQ = 0;
	nZ80IRQCycles = nZ80IRQPeriod;

	memset(nSoundData, 0, sizeof(nSoundData));

	YMZ280BReset();

	ZetRun(4000000 / 60);										// Give the Z80 enough time to initialise

	return 0;
}

static int bbakraidInit()
{
	int nLen;

#ifdef DRIVER_ROTATION
	bToaRotateScreen = true;
#endif

	nGP9001ROMSize[0] = 0x1000000;

	// Find out how much memory is needed
	Mem = NULL;
	MemIndex();
	nLen = MemEnd - (unsigned char *)0;
	if ((Mem = (unsigned char *)malloc(nLen)) == NULL) {
		return 1;
	}
	memset(Mem, 0, nLen);										// blank all memory
	MemIndex();													// Index the allocated memory

	EEPROMInit(4096, 8);										// EEPROM has 4096 bits, uses 8-bit words

	if (strcmp("bbakraid", BurnDrvText(0)) == 0) {
		EEPROMFill(nEEPROMData, 0x0200 - 0x42, 0x42);
	}

	// Load the roms into memory
	if (LoadRoms()) {
		return 1;
	}

	{
		SekInit(1);												// Allocate 68000

		// Map 68000 memory:
	    SekOpen(0);
		SekMapMemory(Rom01,		0x000000, 0x1FFFFF, SM_ROM);	// CPU 0 ROM
		SekMapMemory(Ram02,		0x208000, 0x20FFFF, SM_RAM);

		Map68KTextROM(true);

		SekSetReadWordHandler(0, bbakraidReadWord);
		SekSetReadByteHandler(0, bbakraidReadByte);
		SekSetWriteWordHandler(0, bbakraidWriteWord);
		SekSetWriteByteHandler(0, bbakraidWriteByte);

		SekMapHandler(1,	0x400000, 0x400400, SM_RAM);		// GP9001 addresses

		SekSetReadWordHandler(1, bbakraidReadWordGP9001);
		SekSetWriteWordHandler(1, bbakraidWriteWordGP9001);

		SekMapHandler(2,	0x300000, 0x37FFFF, SM_ROM);		// Z80 ROM

		SekSetReadByteHandler(2, bbakraidReadByteZ80ROM);
		SekSetReadWordHandler(2, bbakraidReadWordZ80ROM);

		SekClose();
	}

	nSpriteYOffset = 0x0001;

	nLayer0XOffset = -0x01D6;
	nLayer1XOffset = -0x01D8;
	nLayer2XOffset = -0x01DA;

	ToaInitGP9001();

	nExtraTXOffset = 0x2C;
	ToaExtraTextInit();

	DrvZ80Init();				// Initialize Z80

	YMZ280BInit(16934400, NULL, 2);

	nToaPalLen = nColCount;
	ToaPalSrc = RamPal;
	ToaPalInit();

	nTextROMStatus = -1;
	bDrawScreen = true;

#if defined _DEBUG && defined USE_SPEEDHACKS
	printf("  * Using speed-hacks (detecting idle loops).\n");
#endif

	DrvDoReset(); 				// Reset machine

	return 0;
}

static int DrvExit()
{
	ToaPalExit();
	YMZ280BExit();
	ToaExitGP9001();
	ToaExtraTextExit();
	ToaZExit();				// Z80 exit
	SekExit();				// Deallocate 68000s

	EEPROMExit();

	// Deallocate all used memory
	free(Mem);
	Mem = NULL;

	return 0;
}

static int DrvDraw()
{
	ToaClearScreen();

	if (bDrawScreen) {
		ToaGetBitmap();
		ToaRenderGP9001();					// Render GP9001 graphics
		ToaExtraTextLayer();				// Render extra text layer
	}

	ToaPalUpdate();							// Update the palette

	return 0;
}

inline static int CheckSleep(int)
{
#if 1 && defined USE_SPEEDHACKS
	int nCurrentPC = SekPc(-1);

	if (!nIRQPending &&
		((nCurrentPC >= 0x000A42 && nCurrentPC <= 0x000A48) ||
		 (nCurrentPC >= 0x001FF0 && nCurrentPC <= 0x001FF6)))
	{
		return 1;
	}
#endif

	return 0;
}

static int DrvFrame()
{
	int nInterleave = 8;

	if (DrvReset) {														// Reset machine
		DrvDoReset();
	}

	// Compile digital inputs
	DrvInput[0] = 0x00;													// Buttons
	DrvInput[1] = 0x00;													// Player 1
	DrvInput[2] = 0x00;													// Player 2
	for (int i = 0; i < 8; i++) {
		DrvInput[0] |= (DrvJoy1[i] & 1) << i;
		DrvInput[1] |= (DrvJoy2[i] & 1) << i;
		DrvInput[2] |= (DrvButton[i] & 1) << i;
	}
	ToaClearOpposites(&DrvInput[0]);
	ToaClearOpposites(&DrvInput[1]);

	nCyclesTotal[0] = (int)((long long)16000000 * nBurnCPUSpeedAdjust / (0x0100 * 60));
	nCyclesTotal[1] = TOA_Z80_SPEED / 60;
	nCyclesDone[0] = nCyclesDone[1] = 0;

	nToaCyclesScanline = nCyclesTotal[0] / 262;
	nToaCyclesDisplayStart = nCyclesTotal[0] - ((nCyclesTotal[0] * (TOA_VBLANK_LINES + 240)) / 262);
	nToaCyclesVBlankStart = nCyclesTotal[0] - ((nCyclesTotal[0] * TOA_VBLANK_LINES) / 262);
	bVBlank = false;

	int nSoundBufferPos = 0;

	SekOpen(0);
	for (int i = 1; i <= nInterleave; i++) {
    	int nCurrentCPU;
		int nNext;

		// Run 68000

		nCurrentCPU = 0;
		nNext = i * nCyclesTotal[nCurrentCPU] / nInterleave;

		// Trigger VBlank interrupt
		if (!bVBlank && nNext > nToaCyclesVBlankStart) {
			if (nCyclesDone[nCurrentCPU] < nToaCyclesVBlankStart) {
				nCyclesSegment = nToaCyclesVBlankStart - nCyclesDone[nCurrentCPU];
				if (!CheckSleep(nCurrentCPU)) {
					nCyclesDone[nCurrentCPU] += SekRun(nCyclesSegment);
				} else {
					nCyclesDone[nCurrentCPU] += nCyclesSegment;
				}
			}

			nIRQPending = 1;
			SekInterrupt(3);
			ToaBufferGP9001Sprites();

			if (pBurnDraw) {
				DrvDraw();												// Draw screen if needed
			}

			bVBlank = true;
		}

		nCyclesSegment = nNext - nCyclesDone[nCurrentCPU];
		if (!CheckSleep(nCurrentCPU)) {									// See if this CPU is busywaiting
			nCyclesDone[nCurrentCPU] += SekRun(nCyclesSegment);
		} else {
			nCyclesDone[nCurrentCPU] += nCyclesSegment;
		}

		if ((i & 1) == 0) {
			// Run Z80
			nCurrentCPU = 1;
			nNext = i * nCyclesTotal[nCurrentCPU] / nInterleave;
			while (nNext >= nZ80IRQCycles) {
				nCyclesSegment = nZ80IRQCycles - nCyclesDone[nCurrentCPU];
				nCyclesDone[nCurrentCPU] += ZetRun(nCyclesSegment);
				ZetRaiseIrq(255);
				nZ80IRQCycles += nZ80IRQPeriod;
			}

			nCyclesSegment = nNext - nCyclesDone[nCurrentCPU];
			if (nCyclesSegment > 0) {
				nCyclesDone[nCurrentCPU] += ZetRun(nCyclesSegment);
			}

			// Render sound segment
			if (pBurnSoundOut) {
				int nSegmentEnd = nBurnSoundLen * i / nInterleave;
				short* pSoundBuf = pBurnSoundOut + (nSoundBufferPos << 1);
				YMZ280BRender(pSoundBuf, nSegmentEnd - nSoundBufferPos);
				nSoundBufferPos = nSegmentEnd;
			}
		}
	}

	nZ80IRQCycles -= nCyclesDone[1];

	SekClose();

	{
		// Make sure the buffer is entirely filled.
		if (pBurnSoundOut) {
			int nSegmentLength = nBurnSoundLen - nSoundBufferPos;
			short* pSoundBuf = pBurnSoundOut + (nSoundBufferPos << 1);
			if (nSegmentLength) {
				YMZ280BRender(pSoundBuf, nSegmentLength);
			}
		}
	}

	return 0;
}

struct BurnDriver BurnDrvBattleBakraidU = {
	{"bbakraid", "Battle Bakraid (ver. Tue Jun 8 1999)", "Unlimited version", "Eighting", "Toaplan GP9001 based", "1999", NULL, NULL},
	BDF_GAME_WORKING | TOA_ROTATE_GRAPHICS_CCW, 2, HARDWARE_TOAPLAN_RAIZING,
	NULL, bbakraidRomInfo, bbakraidRomName, bbakraidInputInfo, bbakraidDIPInfo,
	bbakraidInit, DrvExit, DrvFrame, DrvDraw, DrvScan, &ToaRecalcPalette,
	240, 320, 3, 4
};

struct BurnDriver BurnDrvBattleBakraid = {
	{"bbakrada", "Battle Bakraid (ver. Wed Apr 7 1999)", NULL, "Eighting", "Toaplan GP9001 based", "1999", "bbakraid", NULL},
	BDF_GAME_WORKING | BDF_CLONE | TOA_ROTATE_GRAPHICS_CCW, 2, HARDWARE_TOAPLAN_RAIZING,
	NULL, bbakradaRomInfo, bbakradaRomName, bbakraidInputInfo, bbakraidDIPInfo,
	bbakraidInit, DrvExit, DrvFrame, DrvDraw, DrvScan, &ToaRecalcPalette,
	240, 320, 3, 4
};

