// Z80 (Zed Eight-Ty) Interface
#define EMU_DOZE					// Use Dave's 'Doze' Assembler Z80 emulator

#ifdef EMU_DOZE
 #include "doze.h"
#endif

int ZetInit(int nCount);
void ZetExit();
int ZetOpen(int nCPU);
void ZetClose();
int ZetMemCallback(int nStart,int nEnd,int nMode);
int ZetMemEnd();
int ZetMapArea(int nStart, int nEnd, int nMode, unsigned char *Mem);
int ZetMapArea(int nStart, int nEnd, int nMode, unsigned char *Mem01, unsigned char *Mem02);
int ZetReset();
int ZetPc(int n);
int ZetScan(int nAction);

#ifdef EMU_DOZE
	extern int nDozeRunTotal;		// While inside ZetRun(), the total cycles we should be doing
#endif

inline static int ZetMidCycles()	// Cycles executed so far (valid during ZetRun only)
{
#ifdef EMU_DOZE
	return Doze.nCyclesTotal - Doze.nCyclesLeft;
#else
	return 0;
#endif
}

inline static void ZetRunEnd()
{
#ifdef EMU_DOZE
	Doze.nCyclesTotal -= Doze.nCyclesLeft;
	Doze.nCyclesLeft = 0;
#endif
}

inline static void ZetRaiseIrq(int nVal)
{
#ifdef EMU_DOZE
	Doze.nInterruptLatch = nVal;
#endif
}

inline static void ZetLowerIrq()
{
#ifdef EMU_DOZE
	Doze.nInterruptLatch = -1;
#endif
}

inline static void ZetNmi()
{
#ifdef EMU_DOZE
  DozeNmi();
#endif
}

inline static int ZetRun(int nCycles)
{
#ifdef EMU_DOZE
	int nDid = 0;
	Doze.nCyclesTotal = nCycles;
	Doze.nCyclesLeft = nCycles;

	DozeRun();
	nDid = ZetMidCycles();

	Doze.nCyclesLeft = 0;
	Doze.nCyclesTotal = 0;

	return nDid;
#else
	return 0;
#endif
}

void ZetSetReadHandler(unsigned char (__fastcall *pHandler)(unsigned short));
void ZetSetWriteHandler(void (__fastcall *pHandler)(unsigned short, unsigned char));
void ZetSetInHandler(unsigned char (__fastcall *pHandler)(unsigned short));
void ZetSetOutHandler(void (__fastcall *pHandler)(unsigned short, unsigned char));
