#include "app.h"
#include <richedit.h>

#define MAKE_STRING_2(s) #s
#define MAKE_STRING(s) MAKE_STRING_2(s)

// Print text surrounded by a black outline
static void myDrawText(HDC hDC, RECT* pRect, char* szText, int nSizeDelta, int nColour, UINT uFormat)
{
	char szName[32];
	RECT rect;
	HFONT hFont, myFont;
	TEXTMETRIC myMetric;

	GetTextFace(hDC, 32, szName);
	GetTextMetrics(hDC, &myMetric);
	myFont = CreateFont(myMetric.tmHeight + nSizeDelta, 0, 0, 0, FW_BLACK, 0, 0, 0, 0, 0, 0, 0, ANTIALIASED_QUALITY, szName);
	hFont = (HFONT)SelectObject(hDC, myFont);
	SetBkMode(hDC, TRANSPARENT);

	// Shrink the rect to allow space for the outline
	memcpy(&rect, pRect, sizeof(RECT));
	rect.left += 1;
	rect.top += 1;
	rect.right -= 1;
	rect.bottom -= 1;

	// Print the outline first
	SetTextColor(hDC, 0);
	rect.left -= 1;
	rect.right -= 1;
	rect.top -= 1;
	rect.bottom -= 1;
	DrawText(hDC, szText, -1, &rect, uFormat);
	rect.left += 2;
	rect.right += 2;
	DrawText(hDC, szText, -1, &rect, uFormat);
	rect.top += 2;
	rect.bottom += 2;
	DrawText(hDC, szText, -1, &rect, uFormat);
	rect.left -= 2;
	rect.right -= 2;
	DrawText(hDC, szText, -1, &rect, uFormat);
	rect.left += 1;
	rect.right += 1;
	rect.top -= 1;
	rect.bottom -= 1;

	// Then print the text on top
	SetTextColor(hDC, nColour);
	DrawText(hDC, szText, -1, &rect, uFormat);

	SelectObject(hDC, hFont);
	DeleteObject(myFont);
}

static BOOL CALLBACK AbProc(HWND hDlg ,UINT Msg, WPARAM wParam, LPARAM lParam)
{
	if (Msg == WM_INITDIALOG) {
		WndInMid(hDlg, hScrnWnd);
		return TRUE;
	}

	if (Msg == WM_DRAWITEM) {
		switch (wParam) {
			case IDC_FBA_VER: {
				char szVerString[128];
				if ((nBurnVer & 0xFFFF) > 0x9990) {

					int nVer1 = ((nBurnVer >> 16) & 0xFF) + 1;
					if ((nVer1 & 0x0F) > 9) {
						nVer1 += 6;
					}
					int nVer2 = nVer1 & 0x0F;
					nVer1 >>= 4;

					sprintf(szVerString, "FB Alpha v%s (v%i.%i Release Candidate %i)", szAppBurnVer, nVer1, nVer2, nBurnVer & 0x0F);
				} else {
					if (nBurnVer & 0x00FF) {
						sprintf(szVerString, "FB Alpha v%s (alpha version)", szAppBurnVer);
					} else {
						if (nBurnVer & 0xFF00) {
							sprintf(szVerString, "FB Alpha v%s (beta version)", szAppBurnVer);
						} else {
							sprintf(szVerString, "FB Alpha v%s", szAppBurnVer);
						}
					}
				}

				myDrawText(((LPDRAWITEMSTRUCT)lParam)->hDC, &((LPDRAWITEMSTRUCT)lParam)->rcItem, szVerString, 2, RGB(0xFF, 0xF7, 0xDF), DT_CENTER);

				return TRUE;
			}
			case IDC_SPECIALSTRING: {

#ifdef SPECIALBUILD
				char* szSpecialBuild = MAKE_STRING(SPECIALBUILD);

				myDrawText(((LPDRAWITEMSTRUCT)lParam)->hDC, &((LPDRAWITEMSTRUCT)lParam)->rcItem, szSpecialBuild, -1, RGB(0xFF, 0xCF, 0x7F), DT_CENTER);
#else
				char szBuild[256] = "";

				RECT rect = ((LPDRAWITEMSTRUCT)lParam)->rcItem;
				rect.top += 1;
 #ifdef __GNUC__
				sprintf(szBuild, "built on %s, %s (GCC %i.%i.%i, %s%s optimised)", __TIME__, __DATE__, __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__, MAKE_STRING(CPUTYPE), MMX ? ", MMX" : "");
 #else
  #ifdef _MSC_VER
				sprintf(szBuild, "built on %s, %s (Visual C++ %i.%i, %s%s optimised)", __TIME__, __DATE__, _MSC_VER / 100 - 6, _MSC_VER % 100 / 10, MAKE_STRING(CPUTYPE), MMX ? ", MMX" : "");
  #else
				sprintf(szBuild, "built on %s, %s (unknown compiler, %s%s optimised)", __TIME__, __DATE__, MAKE_STRING(CPUTYPE), MMX ? ", MMX" : "");
  #endif
 #endif
				myDrawText(((LPDRAWITEMSTRUCT)lParam)->hDC, &rect, szBuild, -2, RGB(0xC7, 0xC7, 0xC7), DT_CENTER);
#endif
				return TRUE;
			}
		}
	}

	if (Msg == WM_COMMAND && HIWORD(wParam) == BN_CLICKED && LOWORD(wParam) == IDC_SHOWLICENSE) {
		RECT rect = { 0, 0, 0, 80 };

		SendMessage(hDlg, WM_NEXTDLGCTL, (WPARAM)GetDlgItem(hDlg, IDOK), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_SHOWLICENSE), FALSE);

		SendDlgItemMessage(hDlg, IDC_LICENSE, EM_SETMARGINS, EC_LEFTMARGIN, 4);

		{
			char* pszLicense = NULL;
			int nSize = 1024;

			do {
				nSize <<= 1;
				free(pszLicense);
				pszLicense = (char*)malloc(nSize);
			} while (pszLicense && LoadString(hAppInst, ID_LICENSE, pszLicense, nSize) == nSize - 1);

			if (pszLicense) {
				SendDlgItemMessage(hDlg, IDC_LICENSE, WM_SETTEXT, (WPARAM)0, (LPARAM)pszLicense);
			}

			free(pszLicense);
		}

		MapDialogRect(hDlg, &rect);
		int nSize = rect.bottom;
		GetWindowRect(hDlg, &rect);
		MoveWindow(hDlg, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top + nSize, TRUE);

		WndInMid(hDlg, hScrnWnd);
		return 0;
	}

	if (Msg == WM_CLOSE) {
		EndDialog(hDlg,0);
	}
	if (Msg == WM_COMMAND && HIWORD(wParam) == BN_CLICKED && LOWORD(wParam) == IDOK) {
		SendMessage(hDlg,WM_CLOSE,0,0);
	}
	return 0;
}

int AboutCreate()
{
	DialogBox(hAppInst,MAKEINTRESOURCE(IDD_ABOUT),hScrnWnd,AbProc);
	return 1;
}

