// Definitions for compiling with GCC and Cygwin/MinGW
#ifdef __GNUC__
 #define _WIN32_WINDOWS 0x410
 #define _WIN32_IE 0x0400
 #ifndef STRICT
  #define STRICT
 #endif
#endif

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <commdlg.h>

#ifdef _MSC_VER
 #undef for
 #define for if (0) {} else for

 #pragma warning(disable:4201)
 #pragma warning(disable:4127)
 #pragma warning(disable:4800)
 #pragma warning(disable:4244)

 #include <conio.h>
 #define printf _cprintf
#endif

#include <mmsystem.h>

#define DIRECTINPUT_VERSION 0x0700
#include <dinput.h>

// Additions to the Cygwin/MinGW win32 headers
#ifdef __GNUC__
 #include "gnu_win32.h"
#endif

// Compensate for Windows' need to be different from the rest of the world.
#ifdef _MSC_VER

 #undef strcasecmp
 #define strcasecmp _stricmp

 #undef strncasecmp
 #define strncasecmp _strnicmp

#endif

#ifdef __GNUC__

 #undef _strncmpi
 #define _strncmpi strncasecmp

 #undef _strnicmp
 #define _strnicmp strncasecmp

#endif

#include "burn.h"
#include "interface.h"
#include "resource.h"
#include "title.h"

// Macro for releasing a COM object
#define RELEASE(x) { if ((x)) (x)->Release(); (x) = NULL; }

#define KEY_DOWN(Code) ((GetAsyncKeyState(Code) & 0x8000) ? 1 : 0)
#define BZIP_MAX (4)						// Maximum zip files to search through

// Macros used for handling Window Messages
#define HANDLE_WM_ENTERMENULOOP(hwnd, wParam, lParam, fn)		\
    ((fn)((hwnd), (BOOL)(wParam)), 0L)

#define HANDLE_WM_EXITMENULOOP(hwnd, wParam, lParam, fn)		\
    ((fn)((hwnd), (BOOL)(wParam)), 0L)

#define HANDLE_WM_ENTERSIZEMOVE(hwnd, wParam, lParam, fn)		\
    ((fn)(hwnd), 0L)

#define HANDLE_WM_EXITSIZEMOVE(hwnd, wParam, lParam, fn)		\
    ((fn)(hwnd), 0L)

#define HANDLE_WM_UNINITMENUPOPUP(hwnd,wParam,lParam,fn)		\
	((fn)((hwnd), (HMENU)(wParam), (UINT)LOWORD(lParam), (BOOL)HIWORD(lParam)),0)

// Extra macro used for handling Window Messages
#define HANDLE_MSGB(hwnd, message, fn)							\
    case (message):												\
         HANDLE_##message((hwnd), (wParam), (lParam), (fn));	\
		 break;

// -----------------------------------------------

// main.cpp
extern HINSTANCE hAppInst;					// Application Instance
extern HANDLE hMainThread;					// Handle to the main thread
extern long int nMainThreadID;				// ID of the main thread
extern int nAppThreadPriority;
extern int nAppShowCmd;

extern HACCEL hAccel;

extern int nAppVirtualFps;					// virtual fps

#define EXE_NAME_SIZE 32
extern char szAppExeName[EXE_NAME_SIZE + 1];
extern char szAppBurnVer[16];

extern bool bCmdOptUsed;
extern bool bAlwaysProcessKeyboardInput;

// Used for the load/save dialog in commdlg.h
extern char szChoice[MAX_PATH];				// File chosen by the user
extern OPENFILENAME ofn;

int AppError(char *szText,int bWarning);
int AppMessage(MSG *pMsg);
bool AppProcessKeyboardInput();

// splash.cpp
int SplashCreate();
void SplashDestroy(bool bForce);

// about.cpp
int AboutCreate();

extern int nSplashTime;

// media.cpp
int MediaInit();
int MediaExit(bool bRestart);

// dat.cpp
int write_datfile(int nDatType, FILE* fDat);
int create_datfile(int nDatType);

// misc.cpp
extern int bDoGamma;
extern int bHardwareGammaOnly;
extern double nGamma;

int AppDirectory();
void RegisterExtensions(bool bCreateKeys);
int GetClientScreenRect(HWND hWnd, RECT *pRect);
int SetBurnHighCol(int nDepth);
int WndInMid(HWND hMid, HWND hBase);
char *InputCodeDesc(int c);
void ComputeGammaLUT();

// drv.cpp
extern int bDrvOkay;						// 1 if the Driver has been initted okay, and it's okay to use the BurnDrv functions
extern char szAppRomPaths[4][MAX_PATH];
int DrvInit(int nDrvNum, bool bRestore);
int DrvInitCallback();						// Used when Burn library needs to load a game. DrvInit(nBurnSelect, false)
int DrvExit();

// run.cpp
extern int bRunPause;
extern int kNetGame;
extern int bAltPause;
int RunIdle();
int RunMessageLoop();
int RunReset();
void ToggleLayer(unsigned char thisLayer);

// zipfn.cpp
struct ZipEntry {
	char* szName;
	unsigned int nLen;
	unsigned int nCrc;
};
int ZipOpen(char* szZip);
int ZipClose();
int ZipGetList(struct ZipEntry** pList, int* pnListCount);
int ZipLoadFile(unsigned char* Dest, int nLen, int* pnWrote, int nEntry);

#define ZIP_LIST_FREE(List, nListCount)			\
{												\
	if (List) {									\
		for (int i = 0; i < nListCount; i++) {	\
			if (List[i].szName) {				\
				free(List[i].szName);			\
				List[i].szName = NULL;			\
			}									\
		}			   							\
		free(List);								\
	}											\
	List = NULL;								\
	nListCount = 0;								\
}

// stringset.cpp
class StringSet {
public:
	char* szText;
	int nLen;
	// printf function to add text to the Bzip string
	int __cdecl Add(char* szFormat, ...);
	int Reset();
	StringSet();
	~StringSet();
};

// bzip.cpp
extern char *szBzipName[BZIP_MAX * 4 + 1];	// RomPaths
extern int nBzipError;						// Non-zero if there is a problem with the opened romset
extern StringSet BzipText;					// Text which describes any problems with loading the zip
extern StringSet BzipDetail;				// Text which describes in detail any problems with loading the zip
int BzipOpen(BOOL);
int BzipClose();
int BzipInit();
int BzipExit();

// scrn.cpp
extern HWND hScrnWnd;						// Handle to the screen window
extern HWND hRebar;							// Handle to the Rebar control containing the menu
extern HWND hwndChat;
extern bool bRescanRoms;
extern bool bMenuEnabled;

extern RECT SystemWorkArea;					// The full screen area

extern int nWindowPosX, nWindowPosY;

int ScrnInit();
int ScrnExit(bool bRestart);
int ScrnSize();
int ScrnTitle();
int ActivateChat();
void DeActivateChat();

// menu.cpp
#define UM_DISPLAYPOPUP (WM_USER + 0x0100)
#define UM_CANCELPOPUP (WM_USER + 0x0101)

extern HANDLE hMenuThread;					// Handle to the thread that executes TrackPopupMenuEx
extern DWORD nMenuThreadID;					// ID of the thread that executes TrackPopupMenuEx
extern HWND hMenubar;						// Handle to the Toolbar control comprising the menu
extern HWND hMenuWindow;
extern bool bMenuDisplayed;
extern int nLastMenu;
extern HMENU hMenu;							// Handle to the menu
extern HMENU hMenuPopup;					// Handle to a popup version of the menu
extern int nMenuHeight;
extern int bAutoPause;
extern int nScreenSize;
extern int nWindowSize;

extern bool bModelessMenu;

int MenuCreate();
void MenuDestroy();
int SetMenuPriority();
void MenuUpdate();
void CreateArcaderesItem();
void MenuEnableItems();
bool MenuHandleKeyboard(MSG*);

// dinp.cpp
int DinpInit(HINSTANCE hInst,HWND hPassWnd);
int DinpExit();
void DinpSetCooperativeLevel(bool bExclusive, bool bForeGround);
int DinpStart();							// Call before checking for Input in a frame
int DinpState(int nCode);
int DinpFind(bool CreateBaseline);
int DinpJoyAxis(int i, int nAxis);
int DinpMouseAxis(int nAxis);

// dsp.cpp
int DspInit();
int DspExit();
int DspDo(short* Wave, int nCount);

// -----------------------------------------------

int InputInit();
int InputExit();
int InputMake(bool bCopy);					// Take inputs
int InputTick();

// sel.cpp
extern int bLoadMenuAutoExpand;
extern int nLoadMenuShowX;
int SelDialog();							// Choose a Burn driver

// cona.cpp
extern int nIniVersion;

struct VidPresetData {
	int nWidth;
	int nHeight;
};

extern struct VidPresetData VidPreset[4];

char *LabelCheck(char* s, char* szLabel);
int ConfigAppLoad();
int ConfigAppSave();

// cong.cpp
extern const int nMinVersion;				// Minimum version of application for which input files are valid
extern bool bSaveInputs;
int ConfigGameLoad(bool bOverWrite);		// char* lpszName = NULL
int ConfigGameSave(bool bSave);

#define QUOTE_MAX (128)						// Maximum length of "quoted strings"
int QuoteRead(char** pszQuote, char** pszEnd, char* szSrc);					// Read a quoted string from szSrc and point to the end

// GameInp structure
#include "gameinp.h"

// gami.cpp
#define SKIP_WS(s) for (;;) { if (*s != ' ' && *s != '\t') break; s++; }	// skip whitespace

extern struct GameInp* GameInp;
extern unsigned int nGameInpCount;
extern unsigned int nMacroCount;
extern unsigned int nMaxMacro;

extern int nAnalogSpeed;

extern int nFireButtons;

extern bool bStreetFighterLayout;
extern bool bLeftAltkeyMapped;

int GameInpInit();
int GameInpExit();
char* InpToDesc(struct GameInp* pgi);
char* InpMacroToDesc(struct GameInp* pgi);
void GameInpCheckLeftAlt();
void GameInpCheckMouse();
int GameInpBlank(int bDipSwitch);
int GameInputAutoIni(int nPlayer, char* lpszFile, bool bOverWrite);
int GameInpDefault();
int GameInpWrite(FILE* h);
int GameInpRead(char* szVal, bool bOverWrite);
int GameInpMacroRead(char* szVal, bool bOverWrite);
int GameInpCustomRead(char* szVal, bool bOverWrite);

// Player Default Controls
extern int nPlayerDefaultControls[4];
extern char szPlayerDefaultIni[4][MAX_PATH];

// wave.cpp
extern FILE* WaveLog;						// wave log file
int WaveLogStart();
int WaveLogStop();

// gamc.cpp
int GamcMisc(struct GameInp* pgi, char* szi, int nPlayer);
int GamcAnalogKey(struct GameInp* pgi, char* szi, int nPlayer, int nSlide);
int GamcAnalogJoy(struct GameInp* pgi, char* szi, int nPlayer, int nJoy, int nSlide);
int GamcPlayer(struct GameInp* pgi, char* szi, int nPlayer, int nDevice);
int GamcPlayerHotRod(struct GameInp* pgi, char* szi, int nPlayer, int nFlags, int nSlide);

// inpd.cpp
extern HWND hInpdDlg;						// Handle to the Input Dialog
int InpdUpdate();
int InpdCreate();
int InpdListMake(int bBuild);

// inpcheat.cpp
extern HWND hInpCheatDlg;					// Handle to the Input Dialog
int InpCheatCreate();
int InpCheatListMake(int bBuild);

// inpdipsw.cpp
extern HWND hInpDIPSWDlg;					// Handle to the Input Dialog
void InpDIPSWResetDIPs();
int InpDIPSWCreate();

// inpmacro.cpp
extern HWND hInpMacroDlg;
void InpMacroExit();
int InpMacroCreate(int nInput);

// inps.cpp
extern HWND hInpsDlg;						// Handle to the Input Set Dialog
extern unsigned int nInpsInput;				// The input number we are redefining
int InpsCreate();
int InpsUpdate();

// inpc.cpp
extern HWND hInpcDlg;						// Handle to the Input Constant Dialog
extern unsigned int nInpcInput;				// The input number we are redefining
int InpcCreate();

// stated.cpp
extern int bDrvSaveAll;
int StatedAuto(int bSave);
int StatedLoad(int nSlot = 0);
int StatedSave(int nSlot = 0);

// numdial.cpp
int NumDialCreate(int bDial);
void GammaDialog();
void ScanlineDialog();
void PhosphorDialog();
void ScreenAngleDialog();
void CPUClockDialog();

// sfactd.cpp
int SFactdCreate();

// roms.cpp
extern int* gameAv;
extern bool avOk;
extern int bAvbOnly;
int RomsDirCreate();
int CreateROMInfo();
void FreeROMInfo();

// res.cpp
int ResCreate();

// fba_kaillera.cpp
int KailleraInitInput();
int KailleraGetInput();

// replay.cpp
extern int nReplayStatus;
int RecordInput();
int ReplayInput();
int StartRecord();
int StartReplay(const char* szFileName = NULL);
void StopReplay();

// conc.cpp
int ConfigCheatLoad();

// sshot.cpp
int MakeScreenShot();

// memcard.cpp
extern int nMemoryCardStatus;	// & 1 = file selected, & 2 = inserted

int	MemCardCreate();
int	MemCardSelect();
int	MemCardInsert();
int	MemCardEject();

// progress.cpp
int ProgressUpdateBurner(double fProgress, const char* pszText, bool bAbs);
int ProgressCreate();
int ProgressDestroy();

