#include "app.h"

static bool SkipComma(char** s)
{
	while (**s && **s != ',') {
		(*s)++;
	}

	if (**s == ',') {
		(*s)++;
	}

	if (**s) {
		return true;
	}

	return false;
}

static void CheatError(char* pszFilename, int nLineNumber, CheatInfo* pCheat, char* pszInfo, char* pszLine)
{
	char szText[1024] = "";

	sprintf(szText, "Cheat file %s is malformed.\nPlease remove or repair the file.\n\n", pszFilename);
	if (pCheat) {
		sprintf(szText + strlen(szText), "Parse error at line %i, in cheat \"%s\".\n", nLineNumber, pCheat->szCheatName);
	} else {
		sprintf(szText + strlen(szText), "Parse error at line %i.\n", nLineNumber);
	}

	if (pszInfo) {
		sprintf(szText + strlen(szText), "Problem:\t%s.\n", pszInfo);
	}
	if (pszLine) {
		sprintf(szText + strlen(szText), "Text:\t%s\n", pszLine);
	}

	AppError(szText, 1);
}

static int ConfigParseFile(char* pszFilename)
{
	char szLine[1024];
	char* s;
	char* t;
	int nLen;

	int nLine = 0;
	char nInside = -1;

	CheatInfo* pCurrentCheat = NULL;

	FILE* h = fopen(pszFilename, "rt");
	if (h == NULL) {
		return 1;
	}

	while (1) {
		if (fgets(szLine, sizeof(szLine), h) == NULL) {
			break;
		}

		nLine++;
		nLen = strlen(szLine);

		// Get rid of the linefeed at the end
		while (szLine[nLen - 1] == 0x0A || szLine[nLen - 1] == 0x0D) {
			szLine[nLen - 1] = 0;
			nLen--;
		}

		s = szLine;													// Start parsing

		if (s[0] == '/' && s[1] == '/') {							// Comment
			continue;
		}

		if ((t = LabelCheck(s, "include")) != 0) {					// Include a file
			s = t;

			char szFilename[MAX_PATH] = "";

			// Read name of the cheat file
			char* szQuote = NULL;
			QuoteRead(&szQuote, NULL, s);

			sprintf(szFilename, "cheat\\%s.dat", szQuote);

			if (ConfigParseFile(szFilename)) {
				sprintf(szFilename, "cheat\\%s.ini", szQuote);
				if (ConfigParseFile(szFilename)) {
					CheatError(pszFilename, nLine, NULL, "included file doesn't exist", szLine);
				}
			}

			continue;
		}

		if ((t = LabelCheck(s, "cheat")) != 0) {					// Add new cheat
			s = t;

			// Read cheat name
			char* szQuote = NULL;
			char* szEnd = NULL;

			QuoteRead(&szQuote, &szEnd, s);

			s = szEnd;

			if ((t = LabelCheck(s, "advanced")) != 0) {				// Advanced cheat
				s = t;
			}

			SKIP_WS(s);

			if (nInside == '{') {
				CheatError(pszFilename, nLine, pCurrentCheat, "missing closing bracket", NULL);
				break;
			}
#if 0
			if (*s != '\0' && *s != '{') {
				CheatError(pszFilename, nLine, NULL, "malformed cheat declaration", szLine);
				break;
			}
#endif
			nInside = *s;

			// Link new node into the list
			CheatInfo* pPreviousCheat = pCurrentCheat;
			pCurrentCheat = (CheatInfo*)malloc(sizeof(CheatInfo));
			if (pCheatInfo == NULL) {
				pCheatInfo = pCurrentCheat;
			}

			memset(pCurrentCheat, 0, sizeof(CheatInfo));
			pCurrentCheat->pPrevious = pPreviousCheat;
			if (pPreviousCheat) {
				pPreviousCheat->pNext = pCurrentCheat;
			}

			// Fill in defaults
			pCurrentCheat->nType = 0;								// Default to cheat type 0 (apply each frame)
			pCurrentCheat->nStatus = -1;							// Disable cheat

			memcpy(pCurrentCheat->szCheatName, szQuote, QUOTE_MAX);

			continue;
		}

		if ((t = LabelCheck(s, "type")) != 0) {						// Cheat type
			if (nInside == -1 || pCurrentCheat == NULL) {
				CheatError(pszFilename, nLine, pCurrentCheat, "rogue cheat type", szLine);
				break;
			}
			s = t;

			// Set type
			pCurrentCheat->nType = strtol(s, NULL, 0);

			continue;
		}

		if ((t = LabelCheck(s, "default")) != 0) {					// Default option
			if (nInside == -1 || pCurrentCheat == NULL) {
				CheatError(pszFilename, nLine, pCurrentCheat, "rogue default", szLine);
				break;
			}
			s = t;

			// Set default option
			pCurrentCheat->nDefault = strtol(s, NULL, 0);

			continue;
		}

		int n = strtol(s, &t, 0);
		if (t != s) {				   								// New option

			if (nInside == -1 || pCurrentCheat == NULL) {
				CheatError(pszFilename, nLine, pCurrentCheat, "rogue option", szLine);
				break;
			}

			// Link a new Option structure to the cheat
			if (n < CHEAT_MAX_OPTIONS) {
				s = t;

				// Read option name
				char* szQuote = NULL;
				char* szEnd = NULL;
				if (QuoteRead(&szQuote, &szEnd, s)) {
					CheatError(pszFilename, nLine, pCurrentCheat, "option name omitted", szLine);
					break;
				}
				s = szEnd;

				if (pCurrentCheat->pOption[n] == NULL) {
					pCurrentCheat->pOption[n] = (CheatOption*)malloc(sizeof(CheatOption));
				}
				memset(pCurrentCheat->pOption[n], 0, sizeof(CheatOption));

				memcpy(pCurrentCheat->pOption[n]->szOptionName, szQuote, QUOTE_MAX);

				int nCurrentAddress = 0;
				bool bOK = true;
				while (nCurrentAddress < CHEAT_MAX_ADDRESS) {
					int nCPU = 0, nAddress = 0, nValue = 0;

					if (SkipComma(&s)) {
						nCPU = strtol(s, &t, 0);		// CPU number
						if (t == s) {
							CheatError(pszFilename, nLine, pCurrentCheat, "CPU number omitted", szLine);
							bOK = false;
							break;
						}
						s = t;

						SkipComma(&s);
						nAddress = strtol(s, &t, 0);	// Address
						if (t == s) {
							bOK = false;
							CheatError(pszFilename, nLine, pCurrentCheat, "address omitted", szLine);
							break;
						}
						s = t;

						SkipComma(&s);
						nValue = strtol(s, &t, 0);		// Value
						if (t == s) {
							bOK = false;
							CheatError(pszFilename, nLine, pCurrentCheat, "value omitted", szLine);
							break;
						}
					} else {
						if (nCurrentAddress) {			// Only the first option is allowed no address
							break;
						}
						if (n) {
							bOK = false;
							CheatError(pszFilename, nLine, pCurrentCheat, "CPU / address / value omitted", szLine);
							break;
						}
					}

					pCurrentCheat->pOption[n]->AddressInfo[nCurrentAddress].nCPU = nCPU;
					pCurrentCheat->pOption[n]->AddressInfo[nCurrentAddress].nAddress = nAddress;
					pCurrentCheat->pOption[n]->AddressInfo[nCurrentAddress].nValue = nValue;
					nCurrentAddress++;
				}

				if (!bOK) {
					break;
				}

			}

			continue;
		}

		SKIP_WS(s);
		if (*s == '}') {
			if (nInside != '{') {
				CheatError(pszFilename, nLine, pCurrentCheat, "missing opening bracket", NULL);
				break;
			}

			nInside = -1;
		}

		// Line isn't (part of) a valid cheat
#if 0
		if (*s) {
			CheatError(pszFilename, nLine, NULL, "rogue line", szLine);
			break;
		}
#endif

	}

	if (h) {
		fclose(h);
	}

	return 0;
}

int ConfigCheatLoad()
{
	char szFilename[MAX_PATH] = "";

	sprintf(szFilename, "cheat\\%s.dat", BurnDrvText(0));
	if(ConfigParseFile(szFilename)) {
		sprintf(szFilename, "cheat\\%s.ini", BurnDrvText(0));
		if(ConfigParseFile(szFilename)) {
			return 1;
		}
	}

	if (pCheatInfo) {
		int nCurrentCheat = 0;
		while (CheatEnable(nCurrentCheat, -1) == 0) {
			nCurrentCheat++;
		}

		CheatUpdate();
	}

	return 0;
}


