// Driver Init module
#include "app.h"

int bDrvOkay = 0;		// 1 if the Driver has been initted okay, and it's okay to use the BurnDrv functions

char szAppRomPaths[4][MAX_PATH] = {{"roms\\"}, {""}, {""}, {""}};

static char *szRomName[BZIP_MAX * 4];

static bool bSaveRAM = false;

static void DrvExitBZip()
{
	int i;

	BzipClose();

	for (i = 0; i < BZIP_MAX * 4; i++) {
		// Free the names of each pathnames
		if (szRomName[i] != NULL) {
			free(szRomName[i]);		// Free the strings
		}
		szRomName[i] = NULL;
		szBzipName[i] = NULL;		// Blank BZip list
	}
	szBzipName[i] = NULL;			// Mark end of zip name list
}

static int DrvInitBZip()
{
	int i;

	memset(&szRomName, 0, sizeof(szRomName));

	for (i = 0; i < BZIP_MAX; i++) {
		char *szName;

		if (BurnDrvGetZipName(&szName, i)) {
			break;
		}

		for (int j = 0; j <= 3; j++) {
			// Allocate space for each full pathname
			if ((szRomName[i * 4 + j] = (char *)malloc(strlen(szName) + strlen(szAppRomPaths[j]) + 1)) != NULL) {
				// Use path plus default zip file name
				sprintf (szRomName[i * 4 + j], "%s%s", szAppRomPaths[j], szName);
				// Point bzip to it
				szBzipName[i * 4 + j] = szRomName[i * 4 + j];
			} else {
				continue;
			}
		}
	}
	szBzipName[i * 4] = NULL;		// Mark end of zip name list

	if (i == 0) {
		return 1;
	}

	BzipOpen(FALSE);

	return 0;
}

static int DoLibInit()				// Do Init of Burn library driver
{
	int nRet;

	DrvInitBZip();

	// If there is an error with the romset, report it
	if (nBzipError) {
		char *szTitle;
		int nIcon, nButton;

		// Make the correct title and icon
		if (nBzipError & 1) {
			nIcon = MB_ICONERROR;
			szTitle = APP_TITLE " Error";
		} else {
			nIcon = MB_ICONWARNING;
			szTitle = APP_TITLE " Warning";
		}

		if (nBzipError & 0x08) {
			nButton = MB_OK;		// no data at all - pretty basic!
		} else {
			BzipText.Add("\nWould you like more detailed information?\n(For experts only!)\n");
			nButton = MB_DEFBUTTON2 | MB_YESNO;
		}

		if (MessageBox(hScrnWnd, BzipText.szText, szTitle, nButton | nIcon) == IDYES) {
			// Give the more detailed information string
			MessageBox(hScrnWnd, BzipDetail.szText, szTitle, MB_OK | nIcon);
		}
	}

	ProgressCreate();

	nRet = BurnDrvInit();

	DrvExitBZip();

	ProgressDestroy();

	if (nRet) {
		return 1;
	} else {
		return 0;
	}
}

// Catch calls to BurnLoadRom() once the emulation has started;
// Intialise the zip module before forwarding the call, and exit cleanly.
static int DrvLoadRom(unsigned char* Dest, int* pnWrote, int i)
{
	int nRet;

	DrvInitBZip();

	nRet = BurnExtLoadRom(Dest, pnWrote, i);

	DrvExitBZip();

	BurnExtLoadRom = DrvLoadRom;

	ScrnTitle();

	return nRet;
}

int DrvInit(int nDrvNum, bool bRestore)
{
	DrvExit();						// Make sure exitted
	MediaExit(true);
	MediaInit();

	nBurnDrvSelect = nDrvNum;		// Set the driver number

	// Define nMaxPlayers early; GameInpInit() needs it (normally defined in DoLibInit()).
	nMaxPlayers = BurnDrvGetMaxPlayers();
	GameInpInit();					// Init game input

	ConfigGameLoad(true);
	InputMake(true);

	GameInpDefault();

	if (kNetGame) {
		nBurnCPUSpeedAdjust = 0x0100;
	}

	if (DoLibInit()) {				// Init the Burn library's driver
		char szTemp[512];

		BurnDrvExit();				// Exit the driver

		ScrnTitle();
		sprintf (szTemp, "There was an error starting '%s'.\n", BurnDrvText(1));
		MessageBox(hScrnWnd,szTemp,APP_TITLE " Error", MB_OK | MB_ICONERROR);
		return 1;
	}

	BurnExtLoadRom = DrvLoadRom;

	bDrvOkay = 1;					// Okay to use all BurnDrv functions

	bSaveRAM = false;
	if (kNetGame) {
		KailleraInitInput();
	} else {
		if (bRestore) {
			StatedAuto(0);
			bSaveRAM = true;

			ConfigCheatLoad();
		}
	}

	nBurnLayer = 0xFF;				// show all layers

	// Reset the speed throttling code, so we don't 'jump' after the load
	RunReset();

	PostMessage(NULL, WM_APP + 0, 0, 0);

	return 0;
}

int DrvInitCallback()
{
	return DrvInit(nBurnDrvSelect, false);
}

int DrvExit()
{
	if (bDrvOkay) {
		VidExit();

		InvalidateRect(hScrnWnd, NULL, 1);
		UpdateWindow(hScrnWnd);			// Blank screen window

		DestroyWindow(hInpdDlg);		// Make sure the Input Dialog is exited
		DestroyWindow(hInpDIPSWDlg);	// Make sure the DipSwitch Dialog is exited
		DestroyWindow(hInpCheatDlg);	// Make sure the Cheat Dialog is exited

		if (nBurnDrvSelect < nBurnDrvCount) {

			MemCardEject();				// Eject memory card if present

			if (bSaveRAM) {
				StatedAuto(1);			// Save NV (or full) RAM
				bSaveRAM = false;
			}

			ConfigGameSave(bSaveInputs);

			GameInpExit();				// Exit game input
			BurnDrvExit();				// Exit the driver
		}
	}

	BurnExtLoadRom = NULL;

	bDrvOkay = 0;					// Stop using the BurnDrv functions

	bRunPause = 0;					// Don't pause when exitted

	if (bAudOkay) {
		// Write silence into the sound buffer on exit, and for drivers which don't use pBurnSoundOut
		memset(nAudNextSound, 0, nAudSegLen << 2);
	}

	nBurnDrvSelect = ~0U;			// no driver selected

	return 0;
}
