// Burner Input Set dialog module
#include "app.h"

HWND hInpsDlg = NULL;						// Handle to the Input Set Dialog
static HWND hPush = NULL;					// Handle to the Push button
static HWND hKey = NULL;					// Handle to the text
unsigned int nInpsInput = 0;				// The input number we are redefining
static struct BurnInputInfo bii;			// Info about the input
static int nDlgState = 0;
static int nInputCode = -1;					// If in state 3, code N was nInputCode
static int nCounter = 0;					// Counter of frames since the dialog was made
static struct GameInp* pgi = NULL;			// Current GameInp
static struct GameInp OldInp;				// Old GameInp
static int bOldPush = 0;					// 1 if the push button was pressed last time

static int InpsInit()
{
	char szTitle[128];
	memset(&OldInp, 0, sizeof(OldInp));

	hPush = GetDlgItem(hInpsDlg, IDC_INPS_PUSH);
	hKey = GetDlgItem(hInpsDlg, IDC_INPS_KEY);

	pgi = NULL;
	if (nInpsInput >= nGameInpCount + nMacroCount) {		// input out of range
		return 1;
	}
	pgi = GameInp + nInpsInput;

	OldInp = *pgi;
	bOldPush = 0;

	if (nInpsInput >= nGameInpCount) {						// input out of range
		// Set the dialog title
		sprintf(szTitle, "Press a key for %.100s", pgi->Macro.szName);
	} else {
		memset(&bii,0,sizeof(bii));
		BurnDrvGetInputInfo(&bii, nInpsInput);
		if (bii.szName == NULL) {
			bii.szName = "";
		}

		// Set the dialog title
		sprintf(szTitle, "Press a key for %.100s", bii.szName);
	}
	SetWindowText(hInpsDlg, szTitle);

	nDlgState = 1;
	nInputCode = -1;
	nCounter = 0;

	DinpFind(true);

	return 0;
}

static int InpsExit()
{
	bOldPush = 0;
	if (pgi != NULL) {
		*pgi=OldInp;
	}
	memset(&OldInp, 0, sizeof(OldInp));

	GameInpCheckLeftAlt();

	nDlgState = 0;
	hKey = NULL;
	hPush = NULL;
	hInpsDlg = NULL;

	return 0;
}

static int SetInput(int nCode)
{
	if ((pgi->nInput & GIT_GROUP_MACRO) == 0) {
		if (bii.nType & BIT_GROUP_CONSTANT) {					// Don't change dip switches!
			DestroyWindow(hInpsDlg);
			return 0;
		}

		if ((bii.nType & BIT_GROUP_ANALOG) && (nCode & 0xFF) < 0x10) {			// Analog controls
			if (strcmp(bii.szInfo + 4, "-axis-neg") == 0 || strcmp(bii.szInfo + 4, "-axis-pos") == 0) {
				if (nCode & 1) {
					pgi->nInput = GIT_JOYAXIS_POS;
				} else {
					pgi->nInput = GIT_JOYAXIS_NEG;
				}
				pgi->Input.JoyAxis.nJoy = (nCode & 0x0F00) >> 8;
				pgi->Input.JoyAxis.nAxis = (nCode & 0x0F) >> 1;
			} else {											// Map entire axis
				pgi->nInput = GIT_JOYAXIS_FULL;
				pgi->Input.JoyAxis.nJoy = (nCode & 0x0F00) >> 8;
				pgi->Input.JoyAxis.nAxis = (nCode & 0x0F) >> 1;
			}
		} else {
			pgi->nInput = GIT_SWITCH;
			pgi->Input.Switch.nCode = (unsigned short)nCode;
		}
	} else {
		pgi->Macro.nMode = 0x01;							// Mark macro as in use
		pgi->Macro.Switch.nCode = (unsigned short)nCode;	// Assign switch
	}

	OldInp = *pgi;

	InpdListMake(0);											// Update list with new input type

	return 0;
}

static int InpsPushUpdate()
{
	int nPushState = 0;
	if (pgi == NULL || nInpsInput >= nGameInpCount) {
		return 1;
	}

	// See if the push button is pressed
	nPushState = SendMessage(hPush, BM_GETSTATE, 0, 0);
	if (nPushState & BST_PUSHED) {
		nPushState = 1;
	} else {
		nPushState = 0;
	}

	if (nPushState) {
		switch (OldInp.nType) {
			case BIT_DIGITAL:								// Set digital inputs to 1
				pgi->nInput = GIT_SWITCH;
				pgi->Input.Constant.nConst = 1;
				break;
			case BIT_ANALOG_ABS:							// Max analog inputs
				pgi->nInput = GIT_SWITCH;
				pgi->Input.Constant.nConst = 0xFF;
				break;
			case BIT_DIPSWITCH:								// Set dipswitch block to 0xFF
				pgi->nInput = GIT_SWITCH;
				pgi->Input.Constant.nConst = 0xFF;
				break;
		}
	} else {
		// Change back
		*pgi = OldInp;
	}
	if (nPushState != bOldPush) {							// refresh view
		InpdListMake(0);
	}

	bOldPush = nPushState;

	return 0;
}

int InpsUpdate()
{
	static int nPrevPos, nCount;
	char szTemp[128] = "";
	int nFind = -1;

	if (hInpsDlg == NULL) {										// Don't do anything if the dialog isn't created
		return 1;
	}
	if (GetActiveWindow() != hInpsDlg) {						// Dialog not in foreground
		return 1;
	}
	if (nCounter < 0x100000) {									// advance frames since dialog was created
		nCounter++;
	}
	DinpStart();												// Poll joysticks etc

	InpsPushUpdate();

	if (nDlgState == 1) {										// 1 = Waiting for all keys to be released
		nFind = DinpFind(false);
		if (nFind >= 0) {
			// All keys haven't been released yet
			if (nCounter >= 60 && (nCounter & 15) == 0) {
				// Alert the user that a key is stuck
				sprintf (szTemp, "Waiting for %.80s to be released...", InputCodeDesc(nFind));
				SetWindowText(hKey, szTemp);
			}
			return 0;
		}
		// All keys released
		SetWindowText(hKey, "");
		nDlgState = 2;

		return 0;
	}

	if (nDlgState == 2) {										// 2 = Waiting for key N for the control
		nFind = DinpFind(false);
		if (nFind < 0) {										// Still waiting
			return 0;
		}
		nInputCode = nFind;
		sprintf(szTemp, "%s", InputCodeDesc(nInputCode));
		SetWindowText(hKey, szTemp);

		if (nInputCode & 0x4000 && (nInputCode & 0xFF) < 0x10) {// Joystick axes
			nPrevPos = DinpJoyAxis((nInputCode >> 8) & 0x3F, (nInputCode >> 1) & 0x07);
			nCount = 0;
		}

		nDlgState = 3;

		return 0;
	}

	if (nDlgState == 3) {										// 3 = waiting for key N to be released
		if (nInputCode & 0x4000 && (nInputCode & 0xFF) < 0x10) {// Joystick axes

			int nCode = DinpFind(false);
			int nPos = DinpJoyAxis((nInputCode >> 8) & 0x3F, (nInputCode >> 1) & 0x07);
			int nDelta = nPrevPos - nPos;
			nPrevPos = nPos;

			if (nCode >= 0) {
				sprintf(szTemp, "%s", InputCodeDesc(nCode));
				SetWindowText(hKey, szTemp);
			}

			if (nDelta < -0x0100 || nDelta > 0x0100) {
				nCount = 0;
				return 0;
			}

			nCount++;											// Ensure the axis isn't moving for a while (the delay is 1/5th of a second)
			if (nCount < 3) {
				return 0;
			}

			if (nCode >= 0) {
				nInputCode = nCode;
			}
		} else {
			if (DinpState(nInputCode)) {						// Still waiting for release
				return 0;
			}
		}

		// Key released
		SetInput(nInputCode);

		nDlgState = 0;
		DestroyWindow(hInpsDlg);								// Quit dialog
		return 0;
	}

	return 0;
}

static BOOL CALLBACK DialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM /*lParam*/)
{
	if (Msg == WM_INITDIALOG) {
		hInpsDlg = hDlg;
		InpsInit();
		return 1;
	}
	if (Msg == WM_CLOSE) {
		DestroyWindow(hInpsDlg);
		return 0;
	}
	if (Msg == WM_DESTROY) {
		InpsExit();
		return 0;
	}

	if (Msg == WM_COMMAND) {
		int Id = LOWORD(wParam);
		int Notify = HIWORD(wParam);
		if (Id == IDCANCEL && Notify == BN_CLICKED) {
			SendMessage(hDlg, WM_CLOSE, 0, 0);
			return 0;
		}
	}

	return 0;
}

int InpsCreate()
{
	DestroyWindow(hInpsDlg);					// Make sure exitted
	hInpsDlg = NULL;

	hInpsDlg = CreateDialog(hAppInst, MAKEINTRESOURCE(IDD_INPS), hInpdDlg, DialogProc);
	if (hInpsDlg == NULL) {
		return 1;
	}

	WndInMid(hInpsDlg, hInpdDlg);
	ShowWindow(hInpsDlg, SW_NORMAL);

	return 0;
}
