#include "app.h"
#include <shlobj.h>

HWND hRomsDlg = NULL;

int* gameAv = NULL;
bool avOk = false;
int bAvbOnly = 1;

static char szRomDat[EXE_NAME_SIZE + 16];
static DWORD dwScanThreadId = 0;
static HANDLE hScanThread = NULL;
static int nOldSelect = 0;
static char* szRomName[BZIP_MAX * 4];

static HANDLE hEvent = NULL;

static void CreateRomDatName()
{
	memcpy(szRomDat, "cfg\\", 4);
	memcpy(szRomDat + 4, szAppExeName, strlen(szAppExeName));
	memcpy(szRomDat + 4 + strlen(szAppExeName), ".roms.dat", 10);

	return;
}

//Select Directory Dialog//////////////////////////////////////////////////////////////////////////

static BOOL CALLBACK DefInpProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM)
{
	int var;
	BROWSEINFO bInfo;
	ITEMIDLIST* pItemIDList = NULL;
	char buffer[MAX_PATH];
	static bool chOk;

	switch (Msg) {
		case WM_INITDIALOG:
			chOk = false;

			SetDlgItemText(hDlg, IDC_ROMSDIR_EDIT1, szAppRomPaths[0]);
			SetDlgItemText(hDlg, IDC_ROMSDIR_EDIT2, szAppRomPaths[1]);
			SetDlgItemText(hDlg, IDC_ROMSDIR_EDIT3, szAppRomPaths[2]);
			SetDlgItemText(hDlg, IDC_ROMSDIR_EDIT4, szAppRomPaths[3]);

			WndInMid(hDlg, hScrnWnd);
			SetFocus(hDlg);											// Enable Esc=close
			break;

		case WM_COMMAND:

			if (LOWORD(wParam) == IDOK) {

				for (int i = 0; i < 4; i++) {
					if (GetDlgItemText(hDlg, IDC_ROMSDIR_EDIT1 + i, buffer, sizeof(buffer)) && lstrcmp(szAppRomPaths[i], buffer)) {
						chOk = true;
						lstrcpy(szAppRomPaths[i], buffer);
					}
				}

				SendMessage(hDlg, WM_CLOSE, 0, 0);
				break;
			} else {
				if (LOWORD(wParam) == IDC_ROMSDIR_BR1) {
					var = IDC_ROMSDIR_EDIT1;
				} else {
					if (LOWORD(wParam) == IDC_ROMSDIR_BR2) {
						var = IDC_ROMSDIR_EDIT2;
					} else {
						if (LOWORD(wParam) == IDC_ROMSDIR_BR3) {
							var = IDC_ROMSDIR_EDIT3;
						} else {
							if (LOWORD(wParam) == IDC_ROMSDIR_BR4) {
								var = IDC_ROMSDIR_EDIT4;
							} else {
								if (HIWORD(wParam) == BN_CLICKED && LOWORD(wParam) == IDCANCEL) {
									SendMessage(hDlg, WM_CLOSE, 0, 0);
								}
								break;
							}
						}
					}
				}
			}

			memset(&bInfo, 0, sizeof(bInfo));
			bInfo.hwndOwner = hDlg;
			bInfo.pszDisplayName = buffer;
			bInfo.lpszTitle = "Select Directory:";
			bInfo.ulFlags = BIF_RETURNONLYFSDIRS;

			pItemIDList = SHBrowseForFolder(&bInfo);

			if (pItemIDList) {
				if (SHGetPathFromIDList(pItemIDList, buffer)) {
					int strLen = lstrlen(buffer);
					if (strLen) {
						if (buffer[strLen - 1] != '\\') {
							buffer[strLen] = '\\';
							buffer[strLen + 1] = '\0';
						}
						SetDlgItemText(hDlg, var, buffer);
					}
				}
			}
			break;

		case WM_CLOSE:
			EndDialog(hDlg, 0);
			if (chOk) {
				bRescanRoms = true;
				CreateROMInfo();
			}
	}

	return 0;
}


int RomsDirCreate()
{
	DialogBox(hAppInst, MAKEINTRESOURCE(IDD_ROMSDIR), hScrnWnd, DefInpProc);
	return 1;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//Check Romsets Dialog/////////////////////////////////////////////////////////////////////////////

static int WriteGameAvb()
{
	FILE *h;

	CreateRomDatName();

	if ((h = fopen(szRomDat, "wt")) == NULL) {
		return 1;
	}

	fprintf(h, APP_TITLE " v%.20s ROMs", szAppBurnVer);	// identifier
	fprintf(h, " 0x%04X ", nBurnDrvCount);				// no of games

	for (unsigned int i = 0; i < nBurnDrvCount; i++) {
		if (gameAv[i]) {
			fputc('*', h);
		} else {
			fputc('-', h);
		}
	}

	fprintf(h, " END");									// end marker

	fclose(h);

	return 0;
}

static int DoCheck(char* buffPos)
{
	char label[32];

	// Check identifier
	memset(label, 0, sizeof(label));
	wsprintf(label, APP_TITLE " v%.20s ROMs", szAppBurnVer);
	if ((buffPos = LabelCheck(buffPos, label)) == NULL) {
		return 1;
	}

	// Check no of supported games
	memset(label, 0, sizeof(label));
	memcpy(label, buffPos, 8);
	buffPos += 8;
	unsigned int n = strtol(label, NULL, 0);
	if (n != nBurnDrvCount) {
		return 1;
	}

	for (unsigned int i = 0; i < nBurnDrvCount; i++) {
		if (*buffPos == '*') {
			gameAv[i] = 1;
		} else {
			if (*buffPos == '-') {
				gameAv[i] = 0;
			} else {
				return 1;
			}
		}

		buffPos++;
	}

	memset(label, 0, sizeof(label));
	wsprintf(label, " END");
	if (LabelCheck(buffPos, label) == NULL) {
		avOk = true;
		return 0;
	} else {
		return 1;
	}
}

int CheckGameAvb()
{
	FILE *h;
	int bOK;
	int nBufferSize = nBurnDrvCount + 64;
	char* buffer = (char*)malloc(nBufferSize);

	memset(buffer, 0, nBufferSize);

	CreateRomDatName();

	if ((h = fopen(szRomDat,"r")) == NULL) {
		return 1;
	}

	fgets(buffer, nBufferSize, h);
	fclose(h);
	if (buffer == NULL) {
		return 1;
	}

	bOK = DoCheck(buffer);

	free(buffer);

	return bOK;
}

static int QuitRomsScan()
{
	DWORD dwExitCode;

	GetExitCodeThread(hScanThread, &dwExitCode);

	if (dwExitCode == STILL_ACTIVE) {

		// Signal the scan thread to abort
		SetEvent(hEvent);

		// Wait for the thread to finish
		if (WaitForSingleObject(hScanThread, 15000) != WAIT_OBJECT_0) {
			// If the thread doesn't finish within 15 seconds, forcibly kill it
			TerminateThread(hScanThread, 1);
		}

		CloseHandle(hScanThread);
	}

	CloseHandle(hEvent);

	hEvent = NULL;

	hScanThread = NULL;
	dwScanThreadId = 0;

	BzipClose();

	for (int i = 0;i < BZIP_MAX * 4; i++) {
		// Free the names of each pathnames
		if (szRomName[i]) {
			free(szRomName[i]);						// Free the strings
		}
		szRomName[i] = NULL;
		szBzipName[i] = NULL;						// Blank BZip list
	}

	nBurnDrvSelect = nOldSelect;
	nOldSelect = 0;
	bRescanRoms = false;

	if (avOk) {
		WriteGameAvb();
	}

	return 1;
}

static DWORD WINAPI AnalyzingRoms(LPVOID)					// LPVOID lParam
{
	memset(&szRomName, 0, sizeof(szRomName));

	for (unsigned int z = 0; z < nBurnDrvCount; z++) {
		int i;
		nBurnDrvSelect = z;

		// See if we need to abort
		if (WaitForSingleObject(hEvent, 0) == WAIT_OBJECT_0) {
			ExitThread(0);
		}

		SendDlgItemMessage(hRomsDlg, IDC_WAIT_PROG, PBM_STEPIT, 0, 0);

		for (i = 0; i < BZIP_MAX; i++) {
			char *szName;

			if (BurnDrvGetZipName(&szName, i)) {
				break;
			}

			for (int j = 0; j <= 3; j++) {
				if ((szRomName[i * 4 + j] = (char*)malloc(strlen(szName) + strlen(szAppRomPaths[j]) + 1)) != NULL) {
					sprintf(szRomName[i * 4 + j], "%s%s", szAppRomPaths[j], szName);
					szBzipName[i * 4 + j] = szRomName[i * 4 + j];
				} else {
					continue;
				}
			}
		}

		szBzipName[i * 4] = NULL;					// Mark end of zip name list

		if (i) {
			if (BzipOpen(TRUE))	{
				gameAv[z] = 0;
			} else {
				gameAv[z] = 1;
			}
		} else {
			gameAv[z] = 1;
		}

		for (i = 0; i < BZIP_MAX * 4; i++) {
			// Free the names of each pathnames
			if (szRomName[i]) {
				free(szRomName[i]);					// Free the strings
			}
			szRomName[i] = NULL;
			szBzipName[i] = NULL;					// Blank BZip list
		}
   }

	avOk = true;

	PostMessage(hRomsDlg, WM_CLOSE, 0, 0);

	return 0;
}

static BOOL CALLBACK WaitProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM)		// LPARAM lParam
{
	switch (Msg) {
		case WM_INITDIALOG:
			hRomsDlg = hDlg;
			nOldSelect = nBurnDrvSelect;
			memset(gameAv, 0, sizeof(gameAv));
			SendDlgItemMessage(hDlg, IDC_WAIT_PROG, PBM_SETRANGE, 0, MAKELPARAM(0, nBurnDrvCount));
			SendDlgItemMessage(hDlg, IDC_WAIT_PROG, PBM_SETSTEP, (WPARAM)1, 0);

			ShowWindow(GetDlgItem(hDlg, IDC_WAIT_LABEL_A), TRUE);
			SendMessage(GetDlgItem(hDlg, IDC_WAIT_LABEL_A), WM_SETTEXT, (WPARAM)0, (LPARAM)"Scanning ROMs...");
			ShowWindow(GetDlgItem(hDlg, IDCANCEL), TRUE);

			avOk = false;
			hScanThread = CreateThread(NULL, 0, AnalyzingRoms, NULL, THREAD_TERMINATE, &dwScanThreadId);

			hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

			break;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDCANCEL) {
				PostMessage(hDlg, WM_CLOSE, 0, 0);
			}
			break;

		case WM_CLOSE:
			QuitRomsScan();
			EndDialog(hDlg, 0);
			hRomsDlg = NULL;

	}

	return 0;
}

int CreateROMInfo()
{
	if (gameAv == NULL) {
		gameAv = (int*)malloc(nBurnDrvCount * sizeof(int));
		memset(gameAv, 0, nBurnDrvCount * sizeof(int));
	}

	if (gameAv) {
		if (CheckGameAvb() || bRescanRoms) {
			DialogBox(hAppInst, MAKEINTRESOURCE(IDD_WAIT), hScrnWnd, WaitProc);
		}
	}

	return 1;
}

void FreeROMInfo()
{
	free(gameAv);
	gameAv = NULL;
}
