// Run module
#include "app.h"
#include "kailleraclient.h"

int bRunPause = 0;
int bAltPause = 0;

bool bShowFPS = false;

int kNetGame = 0;							// Non-zero if Kaillera is being used

int counter;								// General purpose variable used when debugging

static unsigned int nNormalLast = 0;		// Last value of timeGetTime()
static int nNormalFrac = 0;					// Extra fraction we did

static int bAppStep = 0, bAppFastForward = 0;
static int bAppDoFast = 0;
static int nFastSpeed = 6;

static int GetInput(bool bCopy)
{
	static int i = 0;
	InputMake(bCopy); 						// get input

	// Update Input dialog ever 3 frames
	if (i == 0) {
		InpdUpdate();
	}

	i++;

	if (i >= 3) {
		i = 0;
	}

	// Update Input Set dialog
	InpsUpdate();
	return 0;
}

static void DisplayFPS()
{
	static time_t fpstimer;
	static unsigned int nPreviousFrames;

	char fpsstring[8];
	time_t temptime = clock();
	float fps = static_cast<float>(nFramesRendered - nPreviousFrames) * CLOCKS_PER_SEC / (temptime - fpstimer);
	sprintf(fpsstring, "%2.1f", fps);
	VidSNewShortMsg(fpsstring, 0xDFDFFF, 480, 0);

	fpstimer = temptime;
	nPreviousFrames = nFramesRendered;
}

// define this function somewhere above RunMessageLoop()
void ToggleLayer(unsigned char thisLayer)
{
	nBurnLayer ^= thisLayer;				// xor with thisLayer
	VidRedraw();
	VidPaint(0);
}

// With or without sound, run one frame.
// If bDraw is true, it's the last frame before we are up to date, and so we should draw the screen
static int RunFrame(int bDraw, int bPause)
{
	static int bOldStretch = 0;
	static int bPrevPause = 0;
	static int bPrevDraw = 0;

	if (bDraw) {
		if (bOldStretch && bVidFullStretch == 0) {	// image may have shrunk
			InvalidateRect(hScrnWnd, NULL, 1);	// blank scrn window
			UpdateWindow(hScrnWnd);
		}
	}

	bOldStretch = bVidFullStretch;

	if (bPrevDraw && !bPause) {
		VidPaint(0);							// paint the screen (no need to validate)
	}

	InputTick();								// Tick one frame of input (for sliders etc)

	if (bPause) {
		GetInput(false);						// Update burner inputs, but not game inputs
		if (bPause != bPrevPause) {
			VidPaint(2);                        // Redraw the screen (to ensure mode indicators are updated)
		}
	} else {

		nFramesEmulated++;
		nCurrentFrame++;

		if (kNetGame) {
			GetInput(true);						// Update inputs
			if (KailleraGetInput()) {			// Synchronize input with Kaillera
				return 0;
			}
		} else {
			if (nReplayStatus == 2) {
				GetInput(false);				// Update burner inputs, but not game inputs
				if (ReplayInput()) {			// Read input from file
					bAltPause = 1;
					bRunPause = 1;
					MenuEnableItems();
					if (!nVidFullscreen) {
						while (ShowCursor(1) <  0) {}
					}
				}
			} else {
				GetInput(true);					// Update inputs
			}
		}

		if (nReplayStatus == 1) {
			RecordInput();						// Write input to file
		}

		if (bDraw) {

			nFramesRendered++;

			if (VidFrame()) {					// Do one frame
				AudBlankSound();
			}

			if (bShowFPS) {
				static unsigned int nPrevDoFPS;
				unsigned int nDoFPS = nFramesRendered % 30;

				if (nPrevDoFPS >= nDoFPS) {
					DisplayFPS();
				}
				nPrevDoFPS = nDoFPS;
			}

		} else {								// frame skipping
			pBurnDraw = NULL;					// Make sure no image is drawn
			if (bDrvOkay) {
				BurnDrvFrame();
			}
		}
	}

	bPrevPause = bPause;
	bPrevDraw = bDraw;

	return 0;
}

// Callback used when DSound needs more sound
static int RunGetNextSound(int bDraw)
{
	if (nAudNextSound == NULL) {
		return 1;
	}

	if (bRunPause && bAppStep == 0) {
		memset(nAudNextSound, 0, nAudSegLen << 2);		// Write silence into the buffer
		RunFrame(bDraw, 1);
		return 0;
	}

	if (bAppDoFast) {									// do more frames
		for (int i = 0; i < nFastSpeed; i++) {
			RunFrame(0, 0);
		}
	}

	// Render frame with sound
	if (bAppStep) {
		pBurnSoundOut = NULL;
	} else {
		pBurnSoundOut = nAudNextSound;
	}
	RunFrame(bDraw, 0);
	if (WaveLog != NULL && pBurnSoundOut != NULL) {		// log to the file
		fwrite(pBurnSoundOut, 1, nBurnSoundLen << 2, WaveLog);
		pBurnSoundOut = NULL;
	}

	bAppStep = 0;										// done one step

	return 0;
}

int RunIdle()
{
	int nTime, nCount;

	if (bAudPlaying) {
		// Run with sound
		AudSoundCheck();
		return 0;
	}

	// Run without sound
	nTime = timeGetTime() - nNormalLast;
	nCount = (nTime * nAppVirtualFps - nNormalFrac) / 100000;
	if (nCount <= 0) {						// No need to do anything for a bit
		Sleep(2);

		return 0;
	}

	nNormalFrac += nCount * 100000;
	nNormalLast += nNormalFrac / nAppVirtualFps;
	nNormalFrac %= nAppVirtualFps;

	if (bAppDoFast){						// Temporarily increase virtual fps
		nCount *= nFastSpeed;
	}
	if (nCount > 100) {						// Limit frame skipping
		nCount = 100;
	}
	if (bRunPause) {
		if (bAppStep) {						// Step one frame
			nCount = 10;
		} else {
			RunFrame(1, 1);					// Paused
			return 0;
		}
	}
	bAppStep = 0;

	for (int i = nCount / 10; i > 0; i--) {	// Mid-frames
		RunFrame(0, 0);
	}
	RunFrame(1, 0);							// End-frame

	return 0;
}

int RunReset()
{
	// Reset the speed throttling code
	nNormalLast = 0; nNormalFrac = 0;
	if (!bAudPlaying) {
		// run without sound
		nNormalLast = timeGetTime();
	}
	return 0;
}

static int RunInit()
{
	// Try to run with sound
	DSoundGetNextSound = RunGetNextSound;
	AudSoundPlay();

	RunReset();

	return 0;
}

static int RunExit()
{
	nNormalLast = 0;
	// Stop sound if it was playing
	AudSoundStop();
	return 0;
}

// The main message loop
int RunMessageLoop()
{
	int bRestartVideo;

	do {
		bRestartVideo = 0;

		if (!bVidOkay) {

			MediaInit();

			// Reinit the video plugin
			VidInit();
			if (!bVidOkay && nVidFullscreen) {

				nVidFullscreen = 0;

				MediaExit(bRestartVideo);
				MediaInit();
				VidInit();
			}
			if (!nVidFullscreen) {
				ScrnSize();
			}

			if (!bVidOkay && (bDrvOkay || bVidUsePlaceholder)) {
				// Maske sure the error will be visible
				SplashDestroy(1);

				AppError("VidInit Failed", 0);
			}

			if (bVidOkay && (bRunPause || !bDrvOkay)) {
				VidRedraw();
			}
		}

		if (nVidFullscreen && bDrvOkay) {												// Hide cursor in fullscreen
			while (ShowCursor(0) >= 0) {}
		} else {
			while (ShowCursor(1) <  0) {}
		}

		RunInit();

		ShowWindow(hScrnWnd, nAppShowCmd);												// Show the screen window
		nAppShowCmd = SW_NORMAL;

		SetForegroundWindow(hScrnWnd);

		if (bDrvOkay && !bRunPause) {
			GameInpCheckLeftAlt();
			GameInpCheckMouse();
		}

		while (1) {
			MSG Msg;
			if (PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE)) {
				// A message is waiting to be processed
				if (Msg.message == WM_QUIT)	{											// Quit program
					break;
				}
				if (Msg.message == (WM_APP + 0)) {										// Restart video
					bRestartVideo = 1;
					break;
				}

				if (bMenuEnabled && nVidFullscreen == 0) {								// Handle keyboard messages for the menu
					if (MenuHandleKeyboard(&Msg)) {
						continue;
					}
				}

				if (Msg.message == WM_SYSKEYDOWN || Msg.message == WM_KEYDOWN) {
					if (Msg.lParam & 0x20000000) {
						// An Alt/AltGr-key was pressed
						switch (Msg.wParam) {
							case 'H': {
								bMenuEnabled = !bMenuEnabled;
								PostMessage(NULL, WM_APP + 0, 0, 0);
								break;
							}
							case 'I':
								if (!kNetGame) {
									bAppFastForward =! bAppFastForward;
								}
								break;
							case 'O':
								if (!kNetGame) {
									bAppStep = 1;
								}
								break;
							case 'P':
								if (bDrvOkay && !kNetGame) {
									bRunPause =! bRunPause;
									bAltPause =! bAltPause;
								} else {
									bRunPause = 0;
									bAltPause = 0;
								}
								MenuUpdate();
								break;
							case '6':
								ToggleLayer(1);
								break;
							case '7':
								ToggleLayer(2);
								break;
							case '8':
								ToggleLayer(4);
								break;
							case '9':
								ToggleLayer(8);
								break;
							case '0':
								nBurnLayer = 0xFF;
								VidRedraw();
								VidPaint(0);
								break;
							case 'U': {
								bShowFPS = !bShowFPS;
								if (bShowFPS) {
									DisplayFPS();
								} else {
									VidSKillShortMsg();
								}
								break;
							}
							case 0xBB: {
								char buffer[15] = "Volume: %lu";
								char *noCtrl = "No Volume Control";

								if (nDSoundVol < 0L) {
									nDSoundVol += 100L;
								}
								if (AudSoundSetVolume() == 0) {
									VidSNewShortMsg(noCtrl);
								} else {
									wsprintf(buffer, buffer, 40L + ((nDSoundVol) / 100L));
									VidSNewShortMsg(buffer);
								}
								break;
							}
							case 0xBD: {
								char buffer[15] = "Volume: %lu";
								char *noCtrl = "No Volume Control";

								if (nDSoundVol > -4000L) {
									nDSoundVol -= 100L;
								}
								if (AudSoundSetVolume() == 0) {
									VidSNewShortMsg(noCtrl);
								} else {
									wsprintf(buffer, buffer, 40L + ((nDSoundVol) / 100L));
									VidSNewShortMsg(buffer);
								}
								break;
							}
							case VK_MENU: {
								continue;
							}
						}
					} else {
						switch (Msg.wParam) {

#ifdef _DEBUG
							case 'N':
								counter--;
								printf("*** New counter value: %04X.\n", counter);
								break;
							case 'M':
								counter++;
								printf("*** New counter value: %04X.\n", counter);
								break;
#endif
							case 'T': {
								if (kNetGame && hwndChat == NULL) {
									if (AppMessage(&Msg)) {
										ActivateChat();
									}
								}
								break;
							}
							case VK_ESCAPE: {
								if (hwndChat) {
									DeActivateChat();
								} else {
									if (bCmdOptUsed) {
										PostQuitMessage(0);
									} else {
										if (nVidFullscreen) {
											nVidFullscreen = 0;
											PostMessage(NULL, WM_APP + 0, 0, 0);
										}
									}
								}
								break;
							}
							case VK_RETURN: {
								if (hwndChat) {
									int i = 0;
									while (EditText[i]) {
										if (EditText[i++] != 0x20) {
											break;
										}
									}
									if (i) {
										kailleraChatSend(EditText);
									}
									DeActivateChat();
								}
								break;
							}
							case VK_F1:
							case VK_F2:
							case VK_F3:
							case VK_F4:
							case VK_F5:
							case VK_F6:
							case VK_F7:
							case VK_F8:
							case VK_F9:
							case VK_F10: {
								// Hotkeys for savestates
								if (GetAsyncKeyState(VK_CONTROL) & 0x80000000) {
									int nSlot = Msg.wParam + 1 - VK_F1;
									if (GetAsyncKeyState(VK_SHIFT) & 0x80000000) {		// Save state
										StatedSave(nSlot);
									} else {											// Load state
										StatedLoad(nSlot);
									}
								}
								break;
							}
						}
					}
				} else {
					if (Msg.message == WM_SYSKEYUP) {
						if (Msg.wParam == VK_MENU) {
							continue;
						}
					}
				}

				// Check for messages for dialogs etc.
				if (AppMessage(&Msg)) {
					if (TranslateAccelerator(hScrnWnd, hAccel, &Msg) == 0) {
						if (hwndChat) {
							TranslateMessage(&Msg);
						}
						DispatchMessage(&Msg);
					}
				}
			} else {
				// No messages are waiting

				bAppDoFast = bAppFastForward;

				if (AppProcessKeyboardInput()) {
					if (KEY_DOWN(VK_F1)) {
						bAppDoFast = 1;
					}
				}

				SplashDestroy(0);
				RunIdle();
			}
		}

		RunExit();

		MediaExit(bRestartVideo);
	} while (bRestartVideo);

	return 0;
}

