// Driver Selector module
// TreeView Version by HyperYagami

// This module uses libpng from http://www.libpng.org

#include "app.h"
#include "png.h"

#ifdef _MSC_VER
	#pragma warning(disable:4611) // interaction between _setjmp and C++ destructors not portable
#endif

// Disable selecting non-available sets
#define DISABLE_NON_AVAILABLE_SELECT 0
// Preview directory
#define PREVIEW_DIRECTORY "previews\\"
// Prompt user on loading non-working sets
#define NON_WORKING_PROMPT_ON_LOAD 1

static HWND hSelDlg = NULL;
static HWND hSelList = NULL;
static int nDialogSelect = 0;									// The driver which this dialog selected

static HBITMAP hBmp = NULL;
static HBITMAP hPreview = NULL;
static HWND hInfoLabel[4] = { NULL, NULL, NULL, NULL };			// 4 things in our Info-box
static HWND hInfoText[4] = {NULL, NULL, NULL, NULL};			// 4 things in our Info-box

static char TreeBuilding = 0;									// if 1, ignore TVN_SELCHANGED messages

#define MASKCPS		(1 << (HARDWARE_PREFIX_CAPCOM	>> 24))
#define MASKNEOGEO	(1 << (HARDWARE_PREFIX_SNK		>> 24))
#define MASKSEGA	(1 << (HARDWARE_PREFIX_SEGA		>> 24))
#define MASKTOAPLAN (1 << (HARDWARE_PREFIX_TOAPLAN	>> 24))
#define MASKCAVE	(1 << (HARDWARE_PREFIX_CAVE		>> 24))
#define MASKMISC	(1 << (HARDWARE_PREFIX_MISC		>> 24))
#define MASKALL		(MASKCPS | MASKNEOGEO | MASKSEGA | MASKTOAPLAN | MASKCAVE)

#define SHOWSHORT	(1 << 16)

int bLoadMenuAutoExpand = 0;
int nLoadMenuShowX = 0;

static int bImageOrientation;

struct NODEINFO {
	int nBurnDrvNo;
	char bIsParent;
	HTREEITEM hTreeHandle;
	char* szRomName;
};

static NODEINFO* nBurnDrv;
static unsigned int nTmpDrvCount;

// Check if a specified driver is working
static bool CheckWorkingStatus(int nDriver)
{
	int nOldnBurnDrvSelect = nBurnDrvSelect;
	nBurnDrvSelect = nDriver;
	bool bStatus = BurnDrvIsWorking();
	nBurnDrvSelect = nOldnBurnDrvSelect;

	return bStatus;
}

static char* MangleGamename(const char* szOldName)
{
	static char szNewName[256] = "";

	char* pszName = szNewName;

	if (nLoadMenuShowX & SHOWSHORT) {
		sprintf(szNewName, "%s  -  ", BurnDrvText(0));
		pszName += strlen(szNewName);
	}

	if (strnicmp(szOldName, "the ", 4) == 0) {
		int x = 0, y = 0;
		while (szOldName[x] && szOldName[x] != '(') {
			x++;
		}
		y = x;
		while (y && szOldName[y - 1] == ' ') {
			y--;
		}
		strncpy(pszName, szOldName + 4, y - 4);
		sprintf(pszName + y - 4, ", the %s", szOldName + x);
	} else {
		strcpy(pszName, szOldName);
	}

	return szNewName;
}

// Make a tree-view control with all drivers
static int SelListMake()
{
	int nOldSelect = nBurnDrvSelect;										// preserve the currently selected driver
	unsigned int i, j;

	free(nBurnDrv);
	nBurnDrv = (NODEINFO*)malloc(nBurnDrvCount * sizeof(NODEINFO));
	memset(nBurnDrv, 0, nBurnDrvCount * sizeof(NODEINFO));

	nTmpDrvCount = 0;

	if (hSelList == NULL) {
		return 1;
	}

	// Add all the driver names to the list

	// 1st: parents
	for (i = 0; i < nBurnDrvCount; i++) {
		TV_INSERTSTRUCT TvItem;

		nBurnDrvSelect = i;													// Switch to driver i

		if (BurnDrvGetFlags() & BDF_BOARDROM) {
			continue;
		}
		if (BurnDrvText(6) != NULL && (BurnDrvGetFlags() & BDF_CLONE)) {	// Skip clones
			continue;
		}
		if (avOk && bAvbOnly && !gameAv[i])	{								// Skip non-available games if needed
			continue;
		}

		int nHardware = 1 << (BurnDrvGetHardwareCode() >> 24);
		if ((nHardware & MASKALL) && (nHardware & nLoadMenuShowX) || ((nHardware & MASKALL) == 0) && (nLoadMenuShowX & (1 << (HARDWARE_PREFIX_MISC >> 8)))) {
			continue;
		}

		memset(&TvItem, 0, sizeof(TvItem));
		TvItem.item.mask = TVIF_TEXT;
		TvItem.hInsertAfter = TVI_SORT;
		TvItem.item.pszText = MangleGamename(BurnDrvText(1));

		nBurnDrv[nTmpDrvCount].nBurnDrvNo = i;
		nBurnDrv[nTmpDrvCount].szRomName = BurnDrvText(0);
		nBurnDrv[nTmpDrvCount].hTreeHandle = (HTREEITEM)SendMessage(hSelList, TVM_INSERTITEM, 0, (LPARAM)&TvItem);
		nBurnDrv[nTmpDrvCount].bIsParent = 1;
		nTmpDrvCount++;
	}

	// 2nd: clones
	for (i = 0; i < nBurnDrvCount; i++) {
		TV_INSERTSTRUCT TvItem;

		nBurnDrvSelect = i;													// Switch to driver i

		if (BurnDrvGetFlags() & BDF_BOARDROM) {
			continue;
		}

		if (BurnDrvText(6) == NULL || !(BurnDrvGetFlags() & BDF_CLONE)) {	// Skip parents
			continue;
		}
		if (avOk && bAvbOnly && !gameAv[i])	{								// Skip non-available games if needed
			continue;
		}

		int nHardware = 1 << (BurnDrvGetHardwareCode() >> 24);
		if ((nHardware & MASKALL) && (nHardware & nLoadMenuShowX) || ((nHardware & MASKALL) == 0) && (nLoadMenuShowX & (1 << (HARDWARE_PREFIX_MISC >> 8)))) {
			continue;
		}

		memset(&TvItem, 0, sizeof(TvItem));
		TvItem.item.mask = TVIF_TEXT;
		TvItem.hInsertAfter = TVI_SORT;
		TvItem.item.pszText = MangleGamename(BurnDrvText(1));

		// Find the parent's handle
		for (j = 0; j < nTmpDrvCount; j++) {
			if (nBurnDrv[j].bIsParent) {
				if (!_stricmp(BurnDrvText(6), nBurnDrv[j].szRomName)) {
					TvItem.hParent = nBurnDrv[j].hTreeHandle;
					break;
				}
			}
		}

		// Find the parent and add a branch to the tree
		if (!TvItem.hParent) {
			char szTempName[9];
			strcpy(szTempName, BurnDrvText(6));
			int nTempBurnDrvSelect = nBurnDrvSelect;
			for (j = 0; j < nBurnDrvCount; j++) {
				nBurnDrvSelect = j;
				if (!strcmp(szTempName, BurnDrvText(0))) {
					TV_INSERTSTRUCT TempTvItem;
					memset(&TempTvItem, 0, sizeof(TempTvItem));
					TempTvItem.item.mask = TVIF_TEXT;
					TempTvItem.hInsertAfter = TVI_SORT;
					TempTvItem.item.pszText = BurnDrvText(1);
					nBurnDrv[nTmpDrvCount].nBurnDrvNo = j;
					nBurnDrv[nTmpDrvCount].szRomName = BurnDrvText(0);
					nBurnDrv[nTmpDrvCount].hTreeHandle = (HTREEITEM)SendMessage(hSelList, TVM_INSERTITEM, 0, (LPARAM)&TempTvItem);
					nBurnDrv[nTmpDrvCount].bIsParent = 1;
					TvItem.hParent = nBurnDrv[nTmpDrvCount].hTreeHandle;
					nTmpDrvCount++;
					break;
				}
			}
			nBurnDrvSelect = nTempBurnDrvSelect;
		}

		nBurnDrv[nTmpDrvCount].nBurnDrvNo = i;
		nBurnDrv[nTmpDrvCount].szRomName = BurnDrvText(0);
		nBurnDrv[nTmpDrvCount].hTreeHandle = (HTREEITEM)SendMessage(hSelList, TVM_INSERTITEM, 0, (LPARAM)&TvItem);
		nTmpDrvCount++;
	}

	for (i = 0; i < nTmpDrvCount; i++) {

		// See if we need to expand the branch of an unavailable or non-working parent
		if (nBurnDrv[i].bIsParent && (bLoadMenuAutoExpand || !gameAv[nBurnDrv[i].nBurnDrvNo] || !CheckWorkingStatus(nBurnDrv[i].nBurnDrvNo))) {
			for (j = 0; j < nTmpDrvCount; j++) {

				// Expand the branch only if a working clone is available
				if (gameAv[nBurnDrv[j].nBurnDrvNo]) {
					nBurnDrvSelect = nBurnDrv[j].nBurnDrvNo;
					if (BurnDrvText(6)) {
						if (strcmp(nBurnDrv[i].szRomName, BurnDrvText(6)) == 0) {
							SendMessage(hSelList, TVM_EXPAND,TVE_EXPAND, (LPARAM)nBurnDrv[i].hTreeHandle);
							break;
						}
					}
				}
			}
		}
	}

	nBurnDrvSelect = nOldSelect;

	return 0;
}

static void MyEndDialog()
{
	SendDlgItemMessage(hSelDlg, IDC_SCREENSHOT_H, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);
	SendDlgItemMessage(hSelDlg, IDC_SCREENSHOT_V, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);
	if (hBmp) {
		DeleteObject((HGDIOBJ)hBmp);
		hBmp = NULL;
	}
	if (hPreview) {
		DeleteObject((HGDIOBJ)hPreview);
		hPreview = NULL;
	}

	EndDialog(hSelDlg, 0);
}

// User clicked ok for a driver in the list
static void SelOkay()
{
	unsigned int nSelect = 0;
	HTREEITEM hSelectHandle = (HTREEITEM)SendMessage(hSelList, TVM_GETNEXTITEM, TVGN_CARET, ~0U);

	if (!hSelectHandle)	{			// Nothing is selected, return without closing the window
		return;
	}

	// Search through nBurnDrv[] for the nBurnDrvNo according to the returned hSelectHandle
	for (unsigned int i = 0; i < nTmpDrvCount; i++) {
		if (hSelectHandle == nBurnDrv[i].hTreeHandle) {
			nSelect = nBurnDrv[i].nBurnDrvNo;
			break;
		}
	}

#if DISABLE_NON_AVAILABLE_SELECT
	if (!gameAv[nSelect]) {			// Game not available, return without closing the window
		return;
	}
#endif

#if NON_WORKING_PROMPT_ON_LOAD
	if (!CheckWorkingStatus(nSelect)) {
		if (MessageBox(hSelDlg, "This game isn't working. Load it anyway?" ,"Warning!", MB_YESNO | MB_DEFBUTTON2 | MB_ICONWARNING) == IDNO) {
			return;
		}
	}
#endif

	nDialogSelect = nSelect;
	MyEndDialog();
}

static void RefreshPanel()
{
	// clear preview shot
	if (hBmp) {
		DeleteObject((HGDIOBJ)hBmp);
		hBmp = NULL;
	}

	SendDlgItemMessage(hSelDlg, IDC_SCREENSHOT_H, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hPreview);
	SendDlgItemMessage(hSelDlg, IDC_SCREENSHOT_V, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);

	// Clear the things in our Info-box
	for (int i = 0; i < 4; i++) {
		SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)"");
		EnableWindow(hInfoLabel[i], FALSE);
	}

	CheckDlgButton(hSelDlg, IDC_CHECKCPS, nLoadMenuShowX & MASKCPS ? BST_UNCHECKED : BST_CHECKED);
	CheckDlgButton(hSelDlg, IDC_CHECKNEOGEO, nLoadMenuShowX & MASKNEOGEO ? BST_UNCHECKED : BST_CHECKED);
	CheckDlgButton(hSelDlg, IDC_CHECKSEGA, nLoadMenuShowX & MASKSEGA ? BST_UNCHECKED : BST_CHECKED);
	CheckDlgButton(hSelDlg, IDC_CHECKTOAPLAN, nLoadMenuShowX & MASKTOAPLAN ? BST_UNCHECKED : BST_CHECKED);
	CheckDlgButton(hSelDlg, IDC_CHECKCAVE, nLoadMenuShowX & MASKCAVE ? BST_UNCHECKED : BST_CHECKED);
	CheckDlgButton(hSelDlg, IDC_CHECKMISC, nLoadMenuShowX & MASKMISC ? BST_UNCHECKED : BST_CHECKED);

	CheckDlgButton(hSelDlg,IDC_CHECKAUTOEXPAND, bLoadMenuAutoExpand ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hSelDlg,IDC_CHECKAVAILABLEONLY, bAvbOnly ? BST_CHECKED : BST_UNCHECKED);
}

static void RebuildEverything()
{
	RefreshPanel();

	TreeBuilding = 1;
	SendMessage(hSelList, WM_SETREDRAW, (WPARAM)FALSE,(LPARAM)TVI_ROOT);		// disable redraw
	SendMessage(hSelList, TVM_DELETEITEM, 0, (LPARAM)TVI_ROOT);					// Destory all nodes
	SelListMake();
	SendMessage(hSelList, WM_SETREDRAW, (WPARAM)TRUE, (LPARAM)TVI_ROOT);		// enable redraw

	// Clear the things in our Info-box
	for (int i = 0; i < 4; i++) {
		SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)"");
		EnableWindow(hInfoLabel[i], FALSE);
	}

	TreeBuilding = 0;
}

#define PNG_SIG_CHECK_BYTES 8
#define PNG_XSIZE_HORI 160
#define PNG_YSIZE_HORI 120
#define PNG_XSIZE_VERT PNG_YSIZE_HORI
#define PNG_YSIZE_VERT PNG_XSIZE_HORI
#define PNG_BITDEPTH_24BIT 24

typedef struct tagIMAGE {
	LONG    width;
	LONG    height;
	DWORD   rowbytes;
	DWORD   imgbytes;
	BYTE**	rowptr;
	BYTE*	bmpbits;
} IMAGE;

static void img_free(IMAGE* img)
{
	free(img->rowptr);
	free(img->bmpbits);
}

static int img_alloc(IMAGE* img)
{
	BYTE *bp, **rp;
	LONG n;

	img->rowbytes = ((DWORD)img->width * PNG_BITDEPTH_24BIT + 31) / 32 * 4;
	img->imgbytes = img->rowbytes * img->height;
	img->rowptr = (BYTE**)malloc((size_t)img->height * sizeof(BYTE*));
	img->bmpbits = (BYTE*)malloc((size_t)img->imgbytes);

	if (img->rowptr == NULL || img->bmpbits == NULL) {
		img_free(img);
		return 0;
	}

	n = img->height;
	rp = img->rowptr;
	bp = img->bmpbits;

	bp += img->imgbytes;
	while (--n >= 0) {
		((DWORD*)bp)[-1] = 0;
		bp -= img->rowbytes;
		*(rp++) = bp;
	}

	return 1;
}

static HBITMAP LoadPNG(HWND hDlg, FILE* fp)
{
	IMAGE img;
	png_uint_32 width, height;
	int bit_depth, color_type;

	// check signature
	unsigned char pngsig[PNG_SIG_CHECK_BYTES];
	fread(pngsig, 1, PNG_SIG_CHECK_BYTES, fp);
	if (png_sig_cmp(pngsig, 0, PNG_SIG_CHECK_BYTES)) {
		return 0;
	}

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) {
		return 0;
	}

	png_infop info_ptr=png_create_info_struct(png_ptr);
	if (!info_ptr) {
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return 0;
	}

	if (setjmp(png_jmpbuf(png_ptr))) {
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		return 0;
	}

	memset(&img, 0, sizeof(IMAGE));
	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, PNG_SIG_CHECK_BYTES);
	png_read_info( png_ptr, info_ptr );
	png_get_IHDR( png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, NULL, NULL, NULL );

	// We need the image to be either 160x120x24 or 120x160x24
	img.width = (LONG)width;
	img.height = (LONG)height;
	if (!((img.width == PNG_XSIZE_HORI && img.height == PNG_YSIZE_HORI) || (img.width == PNG_XSIZE_VERT && img.height == PNG_YSIZE_VERT))) {
		longjmp(png_ptr->jmpbuf, 1);
	}
	if (color_type != PNG_COLOR_TYPE_RGB && color_type != PNG_COLOR_TYPE_RGB_ALPHA) {
		longjmp(png_ptr->jmpbuf, 1);
	}

	// Initialize our img structure
	if (!img_alloc(&img)) {
		longjmp(png_ptr->jmpbuf, 1);
	}

	// If bad things happen in libpng we need to do img_free(&img) as well
	if (setjmp(png_jmpbuf(png_ptr))) {
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		img_free(&img);
		return 0;
	}

	// Read the .PNG image
	png_set_bgr(png_ptr);
	png_read_update_info(png_ptr, info_ptr);
	png_read_image(png_ptr, img.rowptr);
	png_read_end(png_ptr, (png_infop)NULL);
	png_destroy_read_struct( &png_ptr, &info_ptr, (png_infopp)NULL);

	// Create a bitmap for the image
	BITMAPINFO* bi = (BITMAPINFO*)LocalAlloc(LPTR, sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD));
	if (bi == NULL) {
		img_free(&img);
		return 0;
	}

	if (img.width > img.height) {
		bImageOrientation = 0;
	} else {
		bImageOrientation = 1;
	}

	bi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi->bmiHeader.biWidth = img.width;
	bi->bmiHeader.biHeight = img.height;
	bi->bmiHeader.biPlanes = 1;
	bi->bmiHeader.biBitCount = PNG_BITDEPTH_24BIT;
	bi->bmiHeader.biCompression = BI_RGB;
	bi->bmiHeader.biSizeImage = img.imgbytes;
	bi->bmiHeader.biXPelsPerMeter = 0;
	bi->bmiHeader.biYPelsPerMeter = 0;
	bi->bmiHeader.biClrUsed = 0;
	bi->bmiHeader.biClrImportant = 0;

	HDC hDC = GetDC(hDlg);
	BYTE* pBits = NULL;
	HBITMAP hNewBmp = CreateDIBSection(hDC, (BITMAPINFO*)bi, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
	if (pBits) {
		memcpy(pBits, img.bmpbits, img.imgbytes);
	}
	ReleaseDC(hDlg, hDC);
	LocalFree(bi);
	img_free(&img);

	return hNewBmp;
}

static BOOL CALLBACK DialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	if (Msg == WM_INITDIALOG) {
		SendDlgItemMessage(hDlg, IDC_SCREENSHOT_H, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);
		SendDlgItemMessage(hDlg, IDC_SCREENSHOT_V, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);

		hPreview = LoadBitmap(hAppInst, MAKEINTRESOURCE(BMP_PREVIEW));

		hSelDlg = hDlg;
		SetWindowText(hSelDlg,APP_TITLE " - Select Game");
		hSelList = GetDlgItem(hSelDlg, IDC_TREE1);
		hInfoLabel[0] = GetDlgItem(hSelDlg, IDC_LABELROMNAME);
		hInfoLabel[1] = GetDlgItem(hSelDlg, IDC_LABELROMINFO);
		hInfoLabel[2] = GetDlgItem(hSelDlg, IDC_LABELSYSTEM);
		hInfoLabel[3] = GetDlgItem(hSelDlg, IDC_LABELCOMMENT);
		hInfoText[0] = GetDlgItem(hSelDlg, IDC_TEXTROMNAME);
		hInfoText[1] = GetDlgItem(hSelDlg, IDC_TEXTROMINFO);
		hInfoText[2] = GetDlgItem(hSelDlg, IDC_TEXTSYSTEM);
		hInfoText[3] = GetDlgItem(hSelDlg, IDC_TEXTCOMMENT);

#ifndef _DEBUG
		EnableWindow(GetDlgItem(hDlg, IDC_CHECKSEGA), FALSE);
#endif

		RefreshPanel();

		bool bFoundROMs = false;
		for (unsigned int i = 0; i < nBurnDrvCount; i++) {
			if (gameAv[i]) {
				bFoundROMs = true;
				break;
			}
		}
		if (!bFoundROMs) {
			RomsDirCreate();
		}

		TreeBuilding = 1;
		SendMessage(hSelList, WM_SETREDRAW, (WPARAM)FALSE, (LPARAM)TVI_ROOT);
		SelListMake();
		SendMessage(hSelList, WM_SETREDRAW, (WPARAM)TRUE, (LPARAM)TVI_ROOT);
		TreeBuilding = 0;

		WndInMid(hDlg, hScrnWnd);

		return TRUE;
	}

	if (Msg == WM_COMMAND) {
		if (HIWORD(wParam) == BN_CLICKED) {
			int wID = LOWORD(wParam);
			switch (wID) {
				case IDOK:
					SelOkay();
					break;
				case IDROM:
					RomsDirCreate();
					RebuildEverything();
					break;
				case IDRESCAN:
					bRescanRoms = true;
					CreateROMInfo();
					RebuildEverything();
					break;
				case IDCANCEL:
					SendMessage(hDlg, WM_CLOSE, 0, 0);
					return 0;

				case IDC_CHECKCPS:
					nLoadMenuShowX ^= MASKCPS;
					RebuildEverything();
					break;
				case IDC_CHECKNEOGEO:
					nLoadMenuShowX ^= MASKNEOGEO;
					RebuildEverything();
					break;
				case IDC_CHECKSEGA:
					nLoadMenuShowX ^= MASKSEGA;
					RebuildEverything();
					break;
				case IDC_CHECKTOAPLAN:
					nLoadMenuShowX ^= MASKTOAPLAN;
					RebuildEverything();
					break;
				case IDC_CHECKCAVE:
					nLoadMenuShowX ^= MASKCAVE;
					RebuildEverything();
					break;
				case IDC_CHECKMISC:
					nLoadMenuShowX ^= MASKMISC;
					RebuildEverything();
					break;

				case IDC_CHECKAVAILABLEONLY:
					bAvbOnly = !bAvbOnly;
					RebuildEverything();
					break;
				case IDC_CHECKAUTOEXPAND:
					bLoadMenuAutoExpand = !bLoadMenuAutoExpand;
					RebuildEverything();
					break;
			}
		}
	}

	if (Msg == WM_CLOSE) {
		nDialogSelect = -1;
		MyEndDialog();
		return 0;
	}

	if (Msg == WM_NOTIFY) {
		NMHDR* pNmHdr = (NMHDR*)lParam;

		if (pNmHdr->code == NM_DBLCLK) {
			SelOkay();

			// disable double-click node-expand
			SetWindowLong(hSelDlg, DWL_MSGRESULT, 1);

			return 1;
		}

		if (pNmHdr->code == NM_CUSTOMDRAW && LOWORD(wParam) == IDC_TREE1) {
			LPNMLVCUSTOMDRAW lplvcd = (LPNMLVCUSTOMDRAW)lParam;
			HTREEITEM hSelectHandle;
			unsigned int i;
			switch (lplvcd->nmcd.dwDrawStage) {
				case CDDS_PREPAINT: {
					SetWindowLong(hSelDlg, DWL_MSGRESULT, CDRF_NOTIFYITEMDRAW);
					return 1;
				}
				case CDDS_ITEMPREPAINT:	{
					hSelectHandle = (HTREEITEM)(lplvcd->nmcd.dwItemSpec);

					// Gray out the non-selected non-available treeitems
					TV_ITEM TvItem;
					TvItem.hItem = hSelectHandle;
					TvItem.mask = TVIF_STATE;
					SendMessage(hSelList, TVM_GETITEM, 0, (LPARAM)&TvItem);
					if (TvItem.state & TVIS_SELECTED) {
						SetWindowLong(hSelDlg, DWL_MSGRESULT, CDRF_DODEFAULT);
						return 1;
					}

					for (i = 0; i < nTmpDrvCount; i++) {
						if (hSelectHandle == nBurnDrv[i].hTreeHandle) {

							// For parents, change the colour of the background, for clones, change only the text colour
							if (!CheckWorkingStatus(nBurnDrv[i].nBurnDrvNo)) {
								if (nBurnDrv[i].bIsParent) {
									lplvcd->clrText = RGB(0x3F, 0x3F, 0x3F);
									lplvcd->clrTextBk = RGB(0xEF, 0xEF, 0xEF);
								} else {
									lplvcd->clrText = RGB(0xBF, 0xBF, 0xBF);
								}
#if 0
								if (!nBurnDrv[i].bIsParent) {
									break;
								}
#else
								break;
#endif
							}
							if (!gameAv[nBurnDrv[i].nBurnDrvNo]) {
								if (nBurnDrv[i].bIsParent) {
									lplvcd->clrText = RGB(0x00, 0x00, 0x00);
									lplvcd->clrTextBk = RGB(0xFF, 0xCF, 0xCF);
								} else {
									lplvcd->clrText = RGB(0xCB, 0x67, 0x67);
								}
							}
							break;
						}
					}
					SetWindowLong(hSelDlg, DWL_MSGRESULT, CDRF_NEWFONT);
					return 1;
				}
				default: {
					SetWindowLong(hSelDlg, DWL_MSGRESULT, CDRF_DODEFAULT);
					return 1;
				}
			}
		}

		if (pNmHdr->code == TVN_ITEMEXPANDING && !TreeBuilding) {
			SendMessage(hSelList, TVM_SELECTITEM, TVGN_CARET, (LPARAM)((LPNMTREEVIEW)lParam)->itemNew.hItem);
			return FALSE;
		}

		if (pNmHdr->code == TVN_SELCHANGED && !TreeBuilding) {
			int nOldSelect = nBurnDrvSelect;
			HTREEITEM hSelectHandle = (HTREEITEM)SendMessage(hSelList, TVM_GETNEXTITEM, TVGN_CARET, ~0U);

			// Search through nBurnDrv[] for the nBurnDrvNo according to the returned hSelectHandle
			for (unsigned int i = 0; i < nTmpDrvCount; i++) {
				if (hSelectHandle == nBurnDrv[i].hTreeHandle) {
					nBurnDrvSelect = nBurnDrv[i].nBurnDrvNo;
					break;
				}
			}

			// Suck the text out of the drivers via BurnDrvText(i)
			for (int i = 0; i < 4; i++) {
				char szItemText[256];
				szItemText[0] = '\0';

				switch (i) {
					case 0: {
						bool bBracket = false;

						sprintf(szItemText, "%s", BurnDrvText(0));
						if ((BurnDrvGetFlags() & BDF_CLONE) && BurnDrvText(6)) {
							int nOldDrvSelect = nBurnDrvSelect;
							char* pszName = BurnDrvText(6);

							sprintf(szItemText + strlen(szItemText), " (clone of %s", BurnDrvText(6));

							for (nBurnDrvSelect = 0; nBurnDrvSelect < nBurnDrvCount; nBurnDrvSelect++) {
								if (!stricmp(pszName, BurnDrvText(0))) {
									break;
								}
							}
							if (nBurnDrvSelect < nBurnDrvCount) {
								if (BurnDrvText(6)) {
									sprintf(szItemText + strlen(szItemText), ", uses ROMs from %s", BurnDrvText(6));
								}
							}
							nBurnDrvSelect = nOldDrvSelect;
							bBracket = true;
						} else {
							if (BurnDrvText(6)) {
								sprintf(szItemText + strlen(szItemText), "%suses ROMs from %s", bBracket ? ", " : " (", BurnDrvText(6));
								bBracket = true;
							}
						}
						if (bBracket) {
							sprintf(szItemText + strlen(szItemText), ")");
						}
						SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)szItemText);
						EnableWindow(hInfoLabel[i], TRUE);
						break;
					}
					case 1: {
						bool bUseInfo = false;

						if (BurnDrvGetFlags() & BDF_PROTOTYPE) {
							sprintf(szItemText + strlen(szItemText), "prototype");
							bUseInfo = true;
						}
						if (BurnDrvGetFlags() & BDF_BOOTLEG) {
							sprintf(szItemText, "%sbootleg / hack", bUseInfo ? ", " : "");
							bUseInfo = true;
						}
						if (BurnDrvText(7)) {
							sprintf(szItemText + strlen(szItemText), "%suses board-ROMs from %s", bUseInfo ? ", " : "", BurnDrvText(7));
							SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)szItemText);
							EnableWindow(hInfoLabel[i], TRUE);
							bUseInfo = true;
						}
						SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)szItemText);
						EnableWindow(hInfoLabel[i], bUseInfo);
						break;
					}
					case 2: {
						sprintf(szItemText, "%s (%s, %s hardware)", BurnDrvText(3) ? BurnDrvText(3) : "unknown", BurnDrvText(5), BurnDrvText(4));
						SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)szItemText);
						EnableWindow(hInfoLabel[i], TRUE);
						break;
					}
					case 3: {
						SendMessage(hInfoText[i], WM_SETTEXT, (WPARAM)0, (LPARAM)BurnDrvText(2));
						if (BurnDrvText(2) ? strlen(BurnDrvText(2)) : 0) {
							EnableWindow(hInfoLabel[i], TRUE);
						} else {
							EnableWindow(hInfoLabel[i], FALSE);
						}
					}
				}
			}

			if (hBmp) {													// Free the previous preview image
				DeleteObject((HGDIOBJ)hBmp);
				hBmp = NULL;
			}

			// Try to load a .PNG preview image
			char TempName[MAX_PATH];
			strcpy(TempName, PREVIEW_DIRECTORY);
			strcat(TempName, BurnDrvText(0));
			strcat(TempName, ".png");

			FILE* fp = fopen(TempName, "rb");
			if (!fp && BurnDrvText(6)) {								// Try the parent
				strcpy(TempName, PREVIEW_DIRECTORY);
				strcat(TempName, BurnDrvText(6));
				strcat(TempName, ".png");
				fp = fopen(TempName, "rb");
			}
			if (fp) {
				hBmp = LoadPNG(hDlg, fp);
				fclose(fp);
			}

#if 0
			// Try to load a .BMP preview image if no .PNG image loaded
			if (hBmp == NULL) {
				strcpy(TempName, PREVIEW_DIRECTORY);
				strcat(TempName, BurnDrvText(0));
				strcat(TempName, ".bmp");
				hBmp = (HBITMAP)LoadImage(0, TempName, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
				if (!hBmp && BurnDrvText(6)) {							// Try the parent
					strcpy(TempName, PREVIEW_DIRECTORY);
					strcat(TempName, BurnDrvText(6));
					strcat(TempName, ".bmp");
					hBmp = (HBITMAP)LoadImage(0, TempName, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
				}
			}
#endif

			if (hBmp) {													// Show the preview image
				if (bImageOrientation == 0) {
					SendDlgItemMessage(hDlg, IDC_SCREENSHOT_H, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hBmp);
					SendDlgItemMessage(hDlg, IDC_SCREENSHOT_V, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);
				} else {
					SendDlgItemMessage(hDlg, IDC_SCREENSHOT_H, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);
					SendDlgItemMessage(hDlg, IDC_SCREENSHOT_V, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hBmp);
				}
			} else {													// No preview image loaded, show default image
				SendDlgItemMessage(hDlg, IDC_SCREENSHOT_H, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hPreview);
				SendDlgItemMessage(hDlg, IDC_SCREENSHOT_V, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)NULL);
			}

			nBurnDrvSelect = nOldSelect;
		}
	}

	return 0;
}

int SelDialog()
{
	DialogBox(hAppInst, MAKEINTRESOURCE(IDD_SELNEW), hScrnWnd, DialogProc);

	hSelDlg = NULL;
	hSelList = NULL;

	free(nBurnDrv);
	nBurnDrv = NULL;

	return nDialogSelect;
}
