#include "app.h"
#include "png.h"

#define SSHOT_NOERROR 0
#define SSHOT_ERROR_BPP_NOTSUPPORTED 1
#define SSHOT_LIBPNG_ERROR 2
#define SSHOT_OTHER_ERROR 3

#ifdef _MSC_VER
 #pragma warning(disable:4611)			// interaction between _setjmp and C++ destructors not portable
#endif

#define SSHOT_DIRECTORY "sshots\\"

static unsigned char* pSShot = NULL;
static unsigned char* pConvertedImage = NULL;
static png_bytep* pSShotImageRows = NULL;
static FILE* ff;

int MakeScreenShot()
{
    char szSShotName[MAX_PATH];
    int w, h;

    // do our PNG construct things
    png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr) {
		return SSHOT_LIBPNG_ERROR;
	}

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
        png_destroy_write_struct(&png_ptr, (png_infopp)NULL);

		return SSHOT_LIBPNG_ERROR;
    }

    if (setjmp(png_jmpbuf(png_ptr))) {
        png_destroy_write_struct(&png_ptr, &info_ptr);
		free(pConvertedImage);
		pConvertedImage = NULL;

		free(pSShotImageRows);
		pSShotImageRows = NULL;

		fclose(ff);
        remove(szSShotName);

		return SSHOT_LIBPNG_ERROR;
    }

	if (pVidImage == NULL) {
		return SSHOT_OTHER_ERROR;
	}

    if (nVidImageBPP < 2 || nVidImageBPP > 4) {
        return SSHOT_ERROR_BPP_NOTSUPPORTED;
    }

	BurnDrvGetVisibleSize(&w, &h);

	pSShot = pVidImage;

	// Convert the image to 32-bit
	if (nVidImageBPP < 4) {
		unsigned char* pTemp = (unsigned char*)malloc(w * h * sizeof(int));

		if (nVidImageBPP == 2) {
			for (int i = 0; i < h * w; i++) {
				unsigned short nColour = ((unsigned short*)pSShot)[i];

				// Red
		        *(pTemp + i * 4 + 0) = (unsigned char)((nColour & 0x1F) << 3);
			    *(pTemp + i * 4 + 0) |= *(pTemp + 4 * i + 0) >> 5;

				if (nVidImageDepth == 15) {
					// Green
					*(pTemp + i * 4 + 1) = (unsigned char)(((nColour >> 5) & 0x1F) << 3);
					*(pTemp + i * 4 + 1) |= *(pTemp + i * 4 + 1) >> 5;
					// Blue
					*(pTemp + i * 4 + 2) = (unsigned char)(((nColour >> 10)& 0x1F) << 3);
					*(pTemp + i * 4 + 2) |= *(pTemp + i * 4 + 2) >> 5;
				}

				if (nVidImageDepth == 16) {
					// Green
					*(pTemp + i * 4 + 1) = (unsigned char)(((nColour >> 5) & 0x3F) << 2);
					*(pTemp + i * 4 + 1) |= *(pTemp + i * 4 + 1) >> 6;
					// Blue
					*(pTemp + i * 4 + 2) = (unsigned char)(((nColour >> 11) & 0x1F) << 3);
					*(pTemp + i * 4 + 2) |= *(pTemp + i * 4 + 2) >> 5;
				}
			}
        } else {
			memset(pTemp, 0, w * h * sizeof(int));
			for (int i = 0; i < h * w; i++) {
		        *(pTemp + i * 4 + 0) = *(pSShot + i * 3 + 0);
		        *(pTemp + i * 4 + 1) = *(pSShot + i * 3 + 1);
		        *(pTemp + i * 4 + 2) = *(pSShot + i * 3 + 2);
			}
        }

		pConvertedImage = pTemp;

        pSShot = pConvertedImage;
	}

	// Rotate and flip the image
	if (BurnDrvGetFlags() & BDF_ORIENTATION_VERTICAL) {
		unsigned char* pTemp = (unsigned char*)malloc(w * h * sizeof(int));

		for (int x = 0; x < h; x++) {
			if (BurnDrvGetFlags() & BDF_ORIENTATION_FLIPPED) {
				for (int y = 0; y < w; y++) {
					((unsigned int*)pTemp)[(w - y - 1) + x * w] = ((unsigned int*)pSShot)[x + y * h];
				}
			} else {
				for (int y = 0; y < w; y++) {
					((unsigned int*)pTemp)[y + (h - x - 1) * w] = ((unsigned int*)pSShot)[x + y * h];
				}
			}
		}

		free(pConvertedImage);
		pConvertedImage = pTemp;

        pSShot = pConvertedImage;
	}

	// construct our filename -> "romname-mm-dd-hms.png"
    time_t currentTime;
    tm* tmTime;

	time(&currentTime);
    tmTime = localtime(&currentTime);

    sprintf(szSShotName,"%s%s-%.2d-%.2d-%.2d%.2d%.2d.png", SSHOT_DIRECTORY, BurnDrvText(0), tmTime->tm_mon, tmTime->tm_mday, tmTime->tm_hour, tmTime->tm_min, tmTime->tm_sec);

	ff = fopen(szSShotName, "wb");
	if (ff == NULL) {
		png_destroy_write_struct(&png_ptr, &info_ptr);

		free(pConvertedImage);
		pConvertedImage = NULL;

		return SSHOT_OTHER_ERROR;
	}

	png_init_io(png_ptr, ff);

    png_set_IHDR(png_ptr, info_ptr, w, h, 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    png_write_info(png_ptr, info_ptr);

	png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);

    png_set_bgr(png_ptr);

	pSShotImageRows = (png_bytep*)malloc(h * sizeof(png_bytep));
    for (int y = 0; y < h; y++) {
        pSShotImageRows[y] = pSShot + (y * w * sizeof(int));
    }

	png_write_image(png_ptr, pSShotImageRows);
	png_write_end(png_ptr, info_ptr);

	free(pSShotImageRows);
	pSShotImageRows = NULL;

	fclose(ff);

	png_destroy_write_struct(&png_ptr, &info_ptr);

	free(pConvertedImage);
	pConvertedImage = NULL;

	return SSHOT_NOERROR;
}
