// State dialog module
#include "app.h"

int bDrvSaveAll=0;

static int MakeOfn()
{
	memset(&ofn, 0, sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hScrnWnd;
	ofn.lpstrFilter = "FB Alpha Save States (*.fs)\0*.fs\0\0";
	ofn.lpstrFile = szChoice;
	ofn.nMaxFile = sizeof(szChoice);
	ofn.lpstrInitialDir = ".\\state";
	ofn.Flags = OFN_NOCHANGEDIR | OFN_HIDEREADONLY;
	ofn.lpstrDefExt = "fs";
	return 0;
}

// The automatic save
int StatedAuto(int bSave)
{
	static char szName[32] = "";
	int nRet;
	
	sprintf(szName, "cfg\\%.8s.fs", BurnDrvText(0));
	
	if (bSave == 0) {
		nRet = BurnStateLoad(szName, bDrvSaveAll, NULL);		// Load ram
		if (nRet && bDrvSaveAll)	{
			nRet = BurnStateLoad(szName, 0, NULL);				// Couldn't get all - okay just try the nvram
		}
	} else {
		nRet = BurnStateSave(szName, bDrvSaveAll);				// Save ram
	}
	
	return nRet;
}

static void CreateStateName(int nSlot)
{
	sprintf(szChoice, ".\\state\\%s %02x.fs", BurnDrvText(0), nSlot);
}

int StatedLoad(int nSlot)		// int nSlot = 0
{
	int nRet;
	int bOldPause;

	if (nSlot) {
		CreateStateName(nSlot);
	} else {
		if (bDrvOkay) {
			sprintf(szChoice, "%.8s*.fs", BurnDrvText(0));
		} else {
			sprintf(szChoice, "*.fs");
		}
		MakeOfn();
		ofn.lpstrTitle = "Load State";

		bOldPause = bRunPause;
		bRunPause = 1;
		nRet = GetOpenFileName(&ofn);
		bRunPause = bOldPause;

		if (nRet == 0) {		// Error
			return 1;
		}
	}

	nRet = BurnStateLoad(szChoice, 1, &DrvInitCallback);

	// Describe any possible errors:
	if (nRet == 3) {
		AppError("The save state is for the wrong game.", 0);
	} else {
		if (nRet == 4) {
			AppError("This state is too old and cannot be loaded.", 0);
		} else {
			if (nRet == 5) {
				AppError("Emulator is too old to load this state.", 0);
			} else {
				if (nRet) {
					AppError("Error loading state", 0);
				}
			}
		}
	}

	return 0;
}

int StatedSave(int nSlot)		// int nSlot = 0
{
	int bOldPause;

	if (bDrvOkay == 0) {
		return 1;
	}

	if (nSlot) {
		CreateStateName(nSlot);
	} else {
		int nRet;

		sprintf (szChoice, "%.8s", BurnDrvText(0));
		MakeOfn();
		ofn.lpstrTitle = "Save State";
		ofn.Flags |= OFN_OVERWRITEPROMPT;

		bOldPause = bRunPause;
		bRunPause = 1;
		nRet = GetSaveFileName(&ofn);
		bRunPause = bOldPause;

		if (nRet == 0) {		// Error
			return 1;
		}
	}

	if (BurnStateSave(szChoice, 1)) {
		AppError("Error saving state", 1);
	}

	return 0;
}
