#!/usr/bin/perl -w

use strict;

my $Outfile;

# Process command line arguments
if ( $ARGV[0] =~ /^-o/i ) {
   if ( $ARGV[0] =~ /^-o$/i ) {
      $Outfile = $ARGV[1];
   } else {
      $ARGV[0] =~ /(?<=-o)(.*)/i;
      $Outfile = $1;
   }
}

unless ( $Outfile ) {
   die "Usage: $0 -o <output file>\n\n";
}

open( OUTFILE, ">$Outfile" ) or die "\nError: Couldn't open OUTPUT file $Outfile: $!";

print "Generating $Outfile...\n";

# Generate optimized functions

print OUTFILE "#define EIGHTBIT 1\n";
print OUTFILE "#define ROT 0\n\n";
print OUTFILE "#define BPP 16\n";

for ( my $Size = 320; $Size <= 384; $Size += 64 ) {
   print OUTFILE "#define XSIZE $Size\n";

   # No scaling
   print OUTFILE "#define ZOOM 0\n";
   for ( my $Function = 0; $Function < 4; $Function++ ) {
      print OUTFILE "#define ZBUFFER " . $Function . "\n";
      print OUTFILE "#define XFLIP 0\n";
      print OUTFILE "#include \"cave_sprite_render.h\"\n";
      print OUTFILE "#undef XFLIP\n";
      print OUTFILE "#define XFLIP 1\n";
      print OUTFILE "#include \"cave_sprite_render.h\"\n";
      print OUTFILE "#undef XFLIP\n";
      print OUTFILE "#undef ZBUFFER\n";
   }
   print OUTFILE "#undef ZOOM\n\n";

   # Scale up/down
   print OUTFILE "#define XFLIP 0\n";
   for ( my $Function = 0; $Function < 8; $Function++ ) {
      if ( ($Function & 3) == 0 ) {
         print OUTFILE "#define ZOOM " . (1 + ($Function & 4) / 4) . "\n";
      }
      print OUTFILE "#define ZBUFFER " . ($Function & 3) . "\n";
      print OUTFILE "#include \"cave_sprite_render_zoom.h\"\n";
      print OUTFILE "#undef ZBUFFER\n";

      if ( ($Function & 3) == 3 ) {
         print OUTFILE "#undef ZOOM\n\n";
      }
   }
   print OUTFILE "#undef XFLIP\n";
   print OUTFILE "#undef XSIZE\n\n";
}
print OUTFILE "#undef BPP\n\n";
print OUTFILE "#undef ROT\n\n";
print OUTFILE "#undef EIGHTBIT\n\n";

# Create table with functions addresses
print OUTFILE "// Table with all function addresses.\n";
print OUTFILE "static RenderSpriteFunction RenderSpriteFunctionTable[] = {\n";
for ( my $Size = 320; $Size <= 384; $Size += 64 ) {
   my $FunctionName = "&RenderSprite16_" . $Size . "_ROT0";
   for ( my $Function = 0; $Function < 16; $Function++ ) {
      if ( ($Function & 3) == 0) {
         print OUTFILE "\t";
      }
      print OUTFILE "$FunctionName";
      if ( ($Function >= 4) && ($Function < 8) ) {
         print OUTFILE "_FLIPX";
      } else {
         print OUTFILE "_NOFLIP";
      }

      if ( $Function < 8 ) {
         print OUTFILE "_NOZOOM";
      } else {
         if ( ($Function & 4) == 0 ) {
            print OUTFILE "_ZOOMOUT";
         } else {
            print OUTFILE "_ZOOMIN";
         }
      }

      if ( $Function < 8 ) {
         print OUTFILE "_CLIPX";
      } else {
         print OUTFILE "_NOCLIP";
      }

      if ( ($Function & 3) == 0 ) {
         print OUTFILE "_NOZBUFFER";
      } else {
         if ( ($Function & 3) == 1 ) {
            print OUTFILE "_RZBUFFER";
         } else {
            if ( ($Function & 3) == 2 ) {
               print OUTFILE "_WZBUFFER";
            } else {
               print OUTFILE "_RWZBUFFER";
            }
         }
      }

      print OUTFILE "_256";

      print OUTFILE ", ";
      if ( ($Function & 3) == 3) {
         print OUTFILE "\n";
      }
   }
}
print OUTFILE "};\n\n";

print OUTFILE "static RenderSpriteFunction* RenderSprite_ROT0[2] = {\n\t&RenderSpriteFunctionTable[0],\n\t&RenderSpriteFunctionTable[16]\n};\n";

close( OUTFILE );

