#!/usr/bin/perl -w

use strict;

my $Outfile;

# Process command line arguments
if ( $ARGV[0] =~ /^-o/i ) {
	if ( $ARGV[0] =~ /^-o$/i ) {
		$Outfile = $ARGV[1];
	} else {
		$ARGV[0] =~ /(?<=-o)(.*)/i;
		$Outfile = $1;
	}
}

unless ( $Outfile ) {
   die "Usage: $0 -o <output file>\n\n";
}

open( OUTFILE, ">$Outfile" ) or die "\nError: Couldn't open OUTPUT file $Outfile: $!";

print "Generating $Outfile...\n";

# Generate optimized functions

print OUTFILE "#define XFLIP 0\n";
print OUTFILE "#define YFLIP 0\n";
print OUTFILE "#define ROT 0\n\n";
print OUTFILE "#define BPP 16\n";

for ( my $Size = 320; $Size <= 384; $Size += 64 ) {
	print OUTFILE "#define XSIZE $Size\n";
	print OUTFILE "#define EIGHTBIT 1\n";
	for ( my $Function = 0; $Function < 4; $Function++ ) {
		print OUTFILE "#define ROWSCROLL " . ($Function & 1) . "\n";
		print OUTFILE "#define ROWSELECT " . (($Function & 2) / 2) . "\n";
		print OUTFILE "#define DOCLIP 0\n";
		print OUTFILE "#include \"cave_tile_render.h\"\n";
		print OUTFILE "#undef DOCLIP\n";
		print OUTFILE "#define DOCLIP 1\n";
		print OUTFILE "#include \"cave_tile_render.h\"\n";
		print OUTFILE "#undef DOCLIP\n";
		print OUTFILE "#undef ROWSELECT\n";
		print OUTFILE "#undef ROWSCROLL\n";
	}
	print OUTFILE "#undef EIGHTBIT\n\n";
	print OUTFILE "#undef XSIZE\n\n";
}
print OUTFILE "#undef BPP\n\n";
print OUTFILE "#undef ROT\n\n";

print OUTFILE "#undef YFLIP\n";
print OUTFILE "#undef XFLIP\n\n";

# Create table with functions addresses
print OUTFILE "// Table with all function addresses.\n";
print OUTFILE "static RenderTileFunction RenderTileFunctionTable[] = {\n";
for ( my $Size = 320; $Size <= 384; $Size += 64 ) {
   my $FunctionName = "&RenderTile16_" . $Size . "_ROT0_NOFLIP";
   for ( my $Function = 0; $Function < 8; $Function++ ) {
      if ( ($Function & 3) == 0) {
         print OUTFILE "\t";
      }
      print OUTFILE "$FunctionName";
      if ( ($Function & 1) == 0 ) {
         print OUTFILE "_NOROWSCROLL";
      } else {
         print OUTFILE "_ROWSCROLL";
      }
      if ( ($Function & 2) == 0 ) {
         print OUTFILE "_NOROWSELECT";
      } else {
         print OUTFILE "_ROWSELECT";
      }
      if ( ($Function & 4) == 0 ) {
         print OUTFILE "_NOCLIP";
      } else {
         print OUTFILE "_CLIP";
      }
      print OUTFILE "_256";

      print OUTFILE ", ";
      if ( ($Function & 3) == 3) {
         print OUTFILE "\n";
      }
   }
}
print OUTFILE "};\n\n";

print OUTFILE "static RenderTileFunction* RenderTile_ROT0[2] = {\n\t&RenderTileFunctionTable[0],\n\t&RenderTileFunctionTable[8]\n};\n";

close( OUTFILE );

